from album.runner.api import get_args, setup

env_file = """name:  particleSeg3D
channels:
  - pytorch
  - nvidia
  - conda-forge
dependencies:
  - python=3.9
  - gxx
  - cxx-compiler
  - pytorch
  - torchvision
  - torchaudio
  - pytorch-cuda=11.
  - pip
  - numpy~=1.24
  - pip:
    - ParticleSeg3D==0.2.16
"""


# gxx is needed for some pip packages to compile C++ code
# cxx-compiler is needed for some pip packages to compile C++ code
# Note: GeodisTK, a dependency of ParticleSeg3D has a bug in the version 0.1.7 where numpy is missing.
# Hence, an installation of numpy via conda-forge first


def zarr2tiff(load_dir: str, save_dir: str, stack=True) -> None:
    import os
    from pathlib import Path

    import numpy as np
    import zarr
    from tifffile import imwrite
    from tqdm import tqdm

    Path(save_dir).mkdir(parents=True, exist_ok=True)
    image_zarr = zarr.open(load_dir, mode="r")
    num_slices = image_zarr.shape[0]

    if stack:
        # stack all slices in one tiff file
        stacked_tiff = np.stack(
            [image_zarr[i].astype(np.uint16) for i in tqdm(range(num_slices))], axis=0
        )
        imwrite(os.path.join(save_dir, "instance_mask.tiff"), stacked_tiff)
    else:
        for i in tqdm(range(num_slices)):
            imwrite(os.path.join(save_dir, f"{i}.tiff"), image_zarr[i])


def run():
    import os

    os.environ["KMP_DUPLICATE_LIB_OK"] = "True"
    import warnings
    from pathlib import Path

    from particleseg3d.conversion.nifti2zarr import nifti2zarr
    from particleseg3d.inference.inference import predict_cases, setup_model
    from tqdm import tqdm

    args = get_args()

    if os.name == "nt":
        print("Windows detected. Setting number of processes to `0`.")
        args.processes = 0

    images_path = Path(args.input).joinpath("images")  # needs to be images folder

    name = None
    if args.name:
        name = args.name.split()

    if not images_path.exists():
        raise ValueError(
            'The given input path does not exits or has the right content. Must have an "images" folder.'
        )

    if Path(args.output).exists() and any(Path(args.output).iterdir()):
        warnings.warn(
            "The given output path is not empty. This may lead to overwriting existing files."
        )

    if any(images_path.glob("*.nii*")) and not any(images_path.glob("*.zarr")):
        for nii in tqdm(
            Path(images_path).glob("*.nii*"), desc="Converting Nifti to Zarr"
        ):
            nifti2zarr(nii, Path(images_path) / (nii.stem + ".zarr"))

    z_score = tuple(float(num) for num in args.zscore.split())

    trainer, model, config = setup_model(args.model, args.fold.split())
    predict_cases(
        load_dir=args.input,
        save_dir=args.output,
        names=name,
        trainer=trainer,
        model=model,
        config=config,
        target_particle_size=float(args.target_particle_size),
        target_spacing=float(args.target_spacing),
        batch_size=int(args.batch_size),
        processes=int(args.processes),
        min_rel_particle_size=float(args.min_rel_particle_size),
        zscore=z_score,
    )

    # optional: convert output zarr to tiff
    if args.save_tiff:
        for dir in tqdm(Path(args.output).iterdir(), desc="Converting Zarr to Tiff"):
            # avoid replacing segmentation output by input
            if str(dir) == str(Path(args.input).joinpath("images")):
                continue
            for z in dir.glob("*.zarr"):
                zarr2tiff(z, Path(args.save_tiff), stack=True)


setup(
    group="de.mdc",
    name="particleSeg3D-predict",
    version="0.1.0",
    title="particleSeg3D Predict",
    description="A solution to provide the particleSeg3D environment",
    solution_creators=["Jan Philipp Albrecht", "Maximilian Otto"],
    cite=[
        {
            "text": "Scalable, out-of-the box segmentation of individual particles from mineral samples acquired with micro CT, Karol Gotkowski and Shuvam Gupta and Jose R. A. Godinho and Camila G. S. Tochtrop and Klaus H. Maier-Hein and Fabian Isensee, 2023",
            "doi": "10.48550/arXiv.2301.13319",
        }
    ],
    tags=["unet", "machine_learning", "images", "segmentation", "particleSeg3D", "3D"],
    license="Apache v2",
    documentation=["https://github.com/MIC-DKFZ/ParticleSeg3D/tree/main"],
    covers=[],
    album_api_version="0.7.0",
    args=[
        {
            "name": "input",
            "description": "Absolute input path to the base folder that contains the dataset structured in the form of the directories `images` and the metadata.json. In the `images` folder, ZARR of NIFTI data is expected.",
            "type": "string",
            "required": True,
        },
        {
            "name": "output",
            "description": "Absolute output path of the save folder.",
            "type": "string",
            "required": True,
        },
        {
            "name": "model",
            "description": "Absolute path to the directory where the model folder is stored in. If there is no folder with a model found, the pretrained model from ParticleSeg3D gets downloaded. Example: /path/to/model/Task310_particle_seg",
            "type": "string",
            "required": True,
        },
        {
            "name": "name",
            "description": "(Optional) The name(s) without extension (.zarr) of the image(s) that should be used for inference. Multiple names must be separated by spaces.",
            "type": "string",
            "required": False,
        },
        {
            "name": "zscore",
            "description": "(Optional) The target spacing in millimeters given as three numbers separate by spaces.",
            "type": "string",
            "required": False,
            "default": "5850.29762143569 7078.294543817302",
        },
        {
            "name": "target_particle_size",
            "description": "(Optional) The target particle size in pixels given as three numbers separate by spaces.",
            "type": "float",
            "required": False,
            "default": 60,
        },
        {
            "name": "target_spacing",
            "description": "(Optional) The target spacing in millimeters given as three numbers separate by spaces.",
            "type": "float",
            "required": False,
            "default": 0.1,
        },
        {
            "name": "fold",
            "description": "(Optional) The folds to use. 0, 1, 2, 3, 4 or a combination.",
            "type": "string",
            "required": False,
            "default": "0 1 2 3 4",
        },
        {
            "name": "batch_size",
            "description": "(Optional) The batch size to use during each inference iteration. A higher batch size decreases inference time, but increases the required GPU memory.",
            "type": "integer",
            "required": False,
            "default": 6,
        },
        {
            "name": "processes",
            "description": "(Optional) Number of processes to use for parallel processing. Zero to disable multiprocessing.",
            "type": "integer",
            "required": False,
            "default": 2,
        },
        {
            "name": "min_rel_particle_size",
            "description": "(Optional) Minimum relative particle size used for filtering.",
            "type": "float",
            "required": False,
            "default": 0.005,
        },
        {
            "name": "save_tiff",
            "description": "(Optional) Path where to save the predictions as TIFF files. If not set, the TIFF files will not be saved.",
            "type": "string",
            "required": False,
        },
    ],
    run=run,
    dependencies={"environment_file": env_file},
)



"""Module for image operations."""
import numpy as np

from ari3d.gui.ari3d_logging import Ari3dLogger

logger = Ari3dLogger()


def center_crop(image_array, shape):
    """Center crop an image array to the specified shape."""
    d1, d2, d3 = shape

    # check if image already has the correct dimension
    if image_array.shape[:3] == [d1, d2, d3]:
        return image_array, shape

    # check if dimension too large
    if any([d > s for d, s in zip([d1, d2, d3], image_array.shape[:3])]):
        logger.log.error(
            "Crop dimension %s in a dimension larger that input %s"
            % ([d1, d2, d3], image_array.shape[:3])
        )
        return image_array, None

    # center crop image to dimension d1, d2, d3
    x_target_size = np.array([d1, d2, d3])
    offset = tuple((a - b) // 2 for a, b in zip(image_array.shape, x_target_size))
    slices = tuple(slice(o, o + s) for o, s in zip(offset, x_target_size))

    return image_array[slices], shape


def center_pad(image_array, shape):
    """Pad an image array to the specified shape."""
    s1, s2, s3 = shape

    if image_array.shape[:3] == [s1, s2, s3]:
        return image_array, shape

    # check if dimension too small
    if any([d < s for d, s in zip([s1, s2, s3], image_array.shape[:3])]):
        logger.log.error(
            "Dimension of input %s larger than padded to %s"
            % (image_array.shape[:3], [s1, s2, s3])
        )
        return image_array, None

    # center pad image to dimension s1, s2, s3
    pad_left_s1 = (s1 - image_array.shape[0]) // 2
    pad_left_s2 = (s2 - image_array.shape[1]) // 2
    pad_left_s3 = (s3 - image_array.shape[2]) // 2

    pad_right_s1 = s1 - image_array.shape[0] - pad_left_s1
    pad_right_s2 = s2 - image_array.shape[1] - pad_left_s2
    pad_right_s3 = s3 - image_array.shape[2] - pad_left_s3

    image_array = np.pad(
        image_array,
        (
            (pad_left_s1, pad_right_s1),
            (pad_left_s2, pad_right_s2),
            (pad_left_s3, pad_right_s3),
        ),
        mode="constant",
        constant_values=0,
    )
    return image_array, shape



from album.runner.api import get_args, setup

env_file = """name:  feature_extraction
channels:
  - conda-forge
dependencies:
  - python=3.9
  - imagecodecs
  - pip
  - pip:
      - numpy
      - pandas
      - scikit-image
      - anndata
      - joblib
      - scipy
      - nibabel
      - tqdm
"""


def run():
    from pathlib import Path

    import numpy as np
    from io_op import (
        convert_h5ad,
        load_and_process_labelled_image,
        load_non_binary_images,
    )
    from properties import (
        PVE_gradient,
        calculate_properties,
        calculate_surface_areas,
        count_erosions,
    )
    from utils import (
        delete_small_particles,
        erosion_based_on_labels,
        extract_histograms,
        filter_mask_image,
        get_unique_labels,
        replace_extra_rows_with_zeros,
    )

    ### link to inputs
    # Size threshold in number of voxels, default 1000
    size_threshold = get_args().Size_threshold  # 800
    # Defines the number of logical processors used in paralel. -1 is default (uses all)
    # Only change if it crashes.
    numb_threads = get_args().numberTreads  # -1
    # Stepsize for creating mesh (called mesh size in window - resolution of mesh)
    stepsize = get_args().Stepsize  # 1
    # Voxel spacing for creating mesh (keep constant)
    voxel_spacing = (1, 1, 1)
    # Enter angle spacing for calculationg Feret dia
    angle_spacing = get_args().Angle_spacing  # 10
    # voxel size in micron
    voxel_size = get_args().Voxel_size  # 16
    background_mean = get_args().Background_mean  # 6600
    # To load only part of the data
    start_slice = get_args().start_slice  # If loading all data set to: None
    end_slice = get_args().end_slice

    if start_slice == -1:
        start_slice = None

    if end_slice == -1:
        end_slice = None

    #################################################################################
    ############################### Load-Save Paths #################################
    ### it reads tif, tiff, nii.gz and nii

    path = Path(get_args().path)
    # Load binary image with particle mask
    binary_image_path = path.joinpath("mask")
    print(binary_image_path)
    # Load non binary image (grey-scale 16bit)
    non_binary_image_path = path.joinpath("gray")
    print(non_binary_image_path)
    # save geometrical properties
    path_to_save_geometrical_properties = path.joinpath("analysis", "Properties.csv")
    # save inner histograms (inside the particle without the eroded voxels)
    path_to_save_inner_volume_histograms = path.joinpath(
        "analysis", "Inner_histograms.h5ad"
    )
    # save outer (surface layers consisting of all voxels eroded) volume histograms
    path_to_save_outer_volume_histograms = path.joinpath(
        "analysis", "Outer_histograms.h5ad"
    )
    # save bulk histograms (= Inner + Outer)
    path_to_save_bulk_histogram = path.joinpath("analysis", "Bulk_histograms.h5ad")
    # save mesh histograms
    path_to_save_surface_mesh_histograms = path.joinpath(
        "analysis", "Surface_histogram.h5ad"
    )
    # save bulk histograms obtained sfter sobel and smoothening
    path_to_save_bulk_eroded_histogram = path.joinpath(
        "analysis", "Eroded_histograms.h5ad"
    )
    # save gradient
    path_to_save_gradient = path.joinpath("analysis", "Gradient.csv")

    ###################################################################################
    ########################### load the images from stacks ###########################
    print("loading from:", path)

    label_mask, binary_mask = load_and_process_labelled_image(
        binary_image_path, start_slice, end_slice
    )
    print("Label loaded!")

    gray_scale_volume = load_non_binary_images(
        non_binary_image_path, start_slice, end_slice
    )
    print("Grey image loaded!")

    unique_labels = get_unique_labels(label_mask)
    print("Image labeled loaded!")

    ###################################################################################
    ########################### Image processing ######################################

    # prepare label mask, binary mask
    label_mask, binary_mask_rm_small, gray_scale_thresh = delete_small_particles(
        label_mask, binary_mask, gray_scale_volume, size_threshold
    )

    # remove non unique labels
    label_mask_filtered = filter_mask_image(label_mask, unique_labels)
    del label_mask

    unique_labels = np.unique(label_mask_filtered)

    # remove background from unique_labels
    unique_labels = unique_labels[unique_labels != 0]

    # log output
    print("Number of particles after processing:", len(unique_labels))

    # get the surface area of the label mask
    print("Calculating surface areas...")
    surface_areas_df = calculate_surface_areas(
        label_mask_filtered, unique_labels, voxel_spacing, stepsize
    )

    # PVE gradient dataset
    print("Calculating PVE Gradients...")
    surface_properties_mean_intensity, eroded_image_1, eroded_image_2 = PVE_gradient(
        binary_mask_rm_small, gray_scale_thresh, label_mask_filtered, background_mean
    )
    surface_properties_mean_intensity["Ratio"] = surface_properties_mean_intensity[
        "Gradient_2"
    ]
    surface_properties_mean_intensity.to_csv(path_to_save_gradient)

    # count for each row number of erosions of which the gradient is < 0.9 or > 1
    print("Performing optimal number of erosions ...")
    surface_properties_mean_intensity[
        "no_of_erosions"
    ] = surface_properties_mean_intensity.apply(count_erosions, axis=1)
    final_eroded_image = erosion_based_on_labels(
        label_mask_filtered, surface_properties_mean_intensity
    )

    print("Processing finished!")

    #################################################################################
    ############################### HISTOGRAMS ######################################
    # BULK HISTOGRAMS
    print("Starting bulk histogram extraction...")
    bulk_histograms = extract_histograms(
        label_mask_filtered, gray_scale_thresh, numb_threads
    )

    # INNER HISTOGRAM
    print("Starting inner volume histogram extraction...")
    inner_volume_labels = final_eroded_image * label_mask_filtered
    inner_volume_histograms = extract_histograms(
        inner_volume_labels, gray_scale_thresh, numb_threads
    )
    del inner_volume_labels, final_eroded_image

    inner_volume_histograms = replace_extra_rows_with_zeros(
        bulk_histograms, inner_volume_histograms
    )

    # OUTER HISTOGRAMS
    print("Starting outer volume histogram extraction...")
    outer_volume_histograms = bulk_histograms - inner_volume_histograms
    outer_volume_histograms[outer_volume_histograms < 0] = 0
    convert_h5ad(inner_volume_histograms, path_to_save_inner_volume_histograms)
    convert_h5ad(outer_volume_histograms, path_to_save_outer_volume_histograms)

    # SURFACE MESH HISTOGRAM
    # Gets the voxel layer at 1 voxel depth surface mesh and convert/save as h5ad
    binary_surface_mesh_eroded2 = eroded_image_1 - eroded_image_2
    binary_surface_mesh_eroded2_labels = (
        binary_surface_mesh_eroded2 * label_mask_filtered
    )
    del binary_surface_mesh_eroded2
    histograms_surface_mesh = extract_histograms(
        binary_surface_mesh_eroded2_labels, gray_scale_thresh, numb_threads
    )
    histograms_surface_mesh = replace_extra_rows_with_zeros(
        bulk_histograms, histograms_surface_mesh
    )

    # Save the histograms
    convert_h5ad(histograms_surface_mesh, path_to_save_surface_mesh_histograms)
    convert_h5ad(bulk_histograms, path_to_save_bulk_histogram)

    ##################################################################################################
    ############################################# PROPERTIES #########################################
    # List of properties
    # area; #bbox; #bbox_area; #centroid; #convex_image; #coords; #equivalent_diameter; #euler_number; #extent
    # feret_diameter_max; #filled_area; #filled_image; #image; #inertia_tensor; #inertia_tensor_eigvals; #intensity_image
    # label; #local_centroid; #major_axis_length; #max_intensity; #mean_intensity; #min_intensity; #minor_axis_length
    # moments; #moments_central; #moments_normalized; #slice; #solidity; #weighted_centroid; #weighted_local_centroid
    # weighted_moments; #weighted_moments_central; #weighted_moments_normalized

    # From the list above, add to the list the properties to be calculated. Note: area is actually volume

    # ['label', 'area', 'min_intensity', 'max_intensity', 'equivalent_diameter', 'mean_intensity', 'bbox', 'centroid']
    properties = get_args().properties_list.split(",")

    # Call the function with appropriate properties
    properties_bulk = calculate_properties(
        label_mask_filtered, gray_scale_thresh, properties, voxel_size, surface_areas_df
    )
    properties_bulk.to_csv(path_to_save_geometrical_properties)


setup(
    group="de.mdc",
    name="feature_extraction",
    version="0.1.0",
    title="Extracts features, histograms and properties from labelled 3D images",
    description="feature_extraction",
    solution_creators=["Jan Philipp Albrecht", "Jose Ricardo da Assuncao Guadiniho"],
    tags=["interactivity", "workflow"],
    license="MIT",
    documentation=[],
    covers=[],
    album_api_version="0.7.0",
    run=run,
    args=[
        {
            "name": "path",
            "type": "directory",
            "required": True,
            "description": "Path to the directory containing masks and grey images",
        },
        {
            "name": "properties_list",
            "type": "string",
            "required": True,
            "description": "List of properties to be calculated seperated by comma without spaces",
        },
        {
            "name": "Size_threshold",
            "type": "integer",
            "required": False,
            "description": "Size threshold in number of voxels, default 800",
        },
        {
            "name": "numberTreads",
            "type": "integer",
            "required": False,
            "description": "Defines the number of logical processors used in parallel. -1 is default (uses all)",
        },
        {
            "name": "Stepsize",
            "type": "integer",
            "required": False,
            "description": "Stepsize for creating mesh (called mesh size in window - resolution of mesh)",
        },
        {
            "name": "Angle_spacing",
            "type": "integer",
            "required": False,
            "description": "Enter angle spacing for calculating Feret diameter",
        },
        {
            "name": "Voxel_size",
            "type": "integer",
            "required": False,
            "description": "Voxel size in micron",
        },
        {
            "name": "Background_mean",
            "type": "integer",
            "required": False,
            "description": "Background mean value",
        },
        {
            "name": "start_slice",
            "type": "integer",
            "required": False,
            "description": "To load only part of the data, set to None if loading all data",
        },
        {
            "name": "end_slice",
            "type": "integer",
            "required": False,
            "description": "To load only part of the data, set to None if loading all data",
        },
    ],
    dependencies={"environment_file": env_file},
)



"""Controller for managing the help and information features of the ari3d application."""
import webbrowser

from PyQt6.QtWidgets import QMessageBox, QWidget

from ari3d import __version__
from ari3d.resources.default_values import DefaultValues


class HelpController:
    """Controller class for managing help and information features in the ari3d application."""

    def __init__(self, parent: QWidget):
        """Initialize the HelpController with a parent widget."""
        self.parent = parent

    def show_about(self):
        """Show the about dialog with project information."""
        text = f"""
        <html>
        <head/>
        <body>
            <p>{DefaultValues.project_description.value}</p>
            <p><b>Version:</b> {__version__}</p>
            <p><b>License:</b> GNU GENERAL PUBLIC LICENSE v3</p>
            <p><b>Developed by:</b> {DefaultValues.developed_by.value}</p>
        </body>
        </html>
        """
        QMessageBox.about(
            self.parent,
            "ARI3D",
            text,
        )

    def open_code_repository(self):
        """Open the code repository in a web browser."""
        webbrowser.open(DefaultValues.repo_link.value)

    def open_documentation(self):
        """Open the documentation in a web browser."""
        webbrowser.open(DefaultValues.documentation_link.value)

    def open_tutorial(self):
        """Open the tutorial in a web browser."""
        webbrowser.open(DefaultValues.tutorial_link.value)

    def open_article(self):
        """Open the article in a web browser."""
        webbrowser.open(DefaultValues.article_link.value)



"""Controller for managing the extraction solution of projects in ari3d."""
from __future__ import annotations

import os
from datetime import datetime

from PyQt6.QtCore import QThreadPool
from PyQt6.QtWidgets import QDialog, QMessageBox, QWidget
from qtpy import QtCore

from ari3d.gui.ari3d_logging import Ari3dLogger
from ari3d.gui.controller.io.album_c import AlbumController
from ari3d.gui.model.tasks import RunSolutionTask
from ari3d.gui.view.solutions.export_properties_ui import ExportPropertiesDialog


class ExtractPropertiesDialogController:
    """Controller for managing the Extract Properties dialog in ari3d."""

    def _check_project_path(self):
        if self.project_controller.project_path is None:
            raise ValueError("Project path is not set")

    def __init__(
        self,
        parent: QWidget,
        album_api: AlbumController,
        project_controller: OpenProjectController,  # noqa: F821
        interactive: bool = True,
    ):
        """Initialize the Extract Properties dialog controller."""
        self.parent = parent
        self.project_controller = project_controller
        self.output_path = self.project_controller.project_path
        self.album_api = album_api
        self.logger = Ari3dLogger()
        self.extract_histogram = QDialog()
        self.extract_histogram_ui = ExportPropertiesDialog()
        self.interactive = interactive
        self.task = None

        # private attributes
        self._voxel_size = None
        self._background = None
        self._feret_angle = None
        self._mesh_spacing = None
        self._start_slice = None
        self._end_slice = None
        self._remove_small = None
        self._num_threads = None
        self._inertia = None
        self._ferets = None
        self._entropy = None
        self._mean_max = None
        self._aspect_ratio = None
        self._moments = None
        self._euler = None
        self._solidity = None
        self._histograms = None
        self._basic_properties = None
        self._select_all = None

        self.set_defaults_by_parameters()

    def setup_ui(self):
        """Set up the UI for the Extract Properties dialog."""
        self.extract_histogram_ui.setupUi(self.extract_histogram)
        self.extract_histogram.setWindowTitle("Extract Properties")

        self.set_ui_defaults_by_parameters()

        # initialize private attributes with default values from the view
        self._voxel_size = self.extract_histogram_ui.spinBox_5.value()
        self._background = self.extract_histogram_ui.spinBox_6.value()
        self._feret_angle = self.extract_histogram_ui.spinBox_8.value()
        self._mesh_spacing = self.extract_histogram_ui.spinBox_7.value()
        self._start_slice = self.extract_histogram_ui.spinBox_9.value()
        self._end_slice = self.extract_histogram_ui.spinBox_10.value()
        self._remove_small = self.extract_histogram_ui.spinBox_3.value()
        self._num_threads = self.extract_histogram_ui.spinBox_4.value()
        self._inertia = self.extract_histogram_ui.checkBox_34.isChecked()
        self._ferets = self.extract_histogram_ui.checkBox_35.isChecked()
        self._entropy = self.extract_histogram_ui.checkBox_27.isChecked()
        self._mean_max = self.extract_histogram_ui.checkBox_29.isChecked()
        self._aspect_ratio = self.extract_histogram_ui.checkBox_21.isChecked()
        self._moments = self.extract_histogram_ui.checkBox_33.isChecked()
        self._euler = self.extract_histogram_ui.checkBox_36.isChecked()
        self._solidity = self.extract_histogram_ui.checkBox_31.isChecked()

        self.connect_buttons()

    def set_defaults_by_parameters(self):
        """Set the default values for the dialog from the project parameters."""
        if "voxel_size" in self.project_controller.parameters:
            self.voxel_size = self.project_controller.parameters["voxel_size"]

        if "background" in self.project_controller.parameters:
            self.background = self.project_controller.parameters["background"]

        if "feret_angle" in self.project_controller.parameters:
            self.feret_angle = self.project_controller.parameters["feret_angle"]

        if "mesh_spacing" in self.project_controller.parameters:
            self.mesh_spacing = self.project_controller.parameters["mesh_spacing"]

        if "start_slice" in self.project_controller.parameters:
            self.start_slice = self.project_controller.parameters["start_slice"]

        if "end_slice" in self.project_controller.parameters:
            self.end_slice = self.project_controller.parameters["end_slice"]

        if "remove_small" in self.project_controller.parameters:
            self.remove_small = self.project_controller.parameters["remove_small"]

        if "num_threads" in self.project_controller.parameters:
            self.num_threads = self.project_controller.parameters["num_threads"]

        if "inertia" in self.project_controller.parameters:
            self.inertia = self.project_controller.parameters["inertia"]

        if "ferets" in self.project_controller.parameters:
            self.ferets = self.project_controller.parameters["ferets"]

        if "entropy" in self.project_controller.parameters:
            self.entropy = self.project_controller.parameters["entropy"]

        if "mean_max" in self.project_controller.parameters:
            self.mean_max = self.project_controller.parameters["mean_max"]

        if "aspect_ratio" in self.project_controller.parameters:
            self.aspect_ratio = self.project_controller.parameters["aspect_ratio"]

        if "moments" in self.project_controller.parameters:
            self.moments = self.project_controller.parameters["moments"]

        if "euler" in self.project_controller.parameters:
            self.euler = self.project_controller.parameters["euler"]

        if "solidity" in self.project_controller.parameters:
            self.solidity = self.project_controller.parameters["solidity"]

        if "histograms" in self.project_controller.parameters:
            self.histograms = self.project_controller.parameters["histograms"]

        if "basic_properties" in self.project_controller.parameters:
            self.basic_properties = self.project_controller.parameters[
                "basic_properties"
            ]

        if "select_all" in self.project_controller.parameters:
            self.select_all = self.project_controller.parameters["select_all"]

    def set_ui_defaults_by_parameters(self):
        """Set the default values for the dialog UI from the project parameters."""
        if "voxel_size" in self.project_controller.parameters:
            self.extract_histogram_ui.spinBox_5.setValue(
                self.project_controller.parameters["voxel_size"]
            )

        if "background" in self.project_controller.parameters:
            self.extract_histogram_ui.spinBox_6.setValue(
                self.project_controller.parameters["background"]
            )

        if "feret_angle" in self.project_controller.parameters:
            self.extract_histogram_ui.spinBox_8.setValue(
                self.project_controller.parameters["feret_angle"]
            )

        if "mesh_spacing" in self.project_controller.parameters:
            self.extract_histogram_ui.spinBox_7.setValue(
                self.project_controller.parameters["mesh_spacing"]
            )

        if "start_slice" in self.project_controller.parameters:
            self.extract_histogram_ui.spinBox_9.setValue(
                self.project_controller.parameters["start_slice"]
            )

        if "end_slice" in self.project_controller.parameters:
            self.extract_histogram_ui.spinBox_10.setValue(
                self.project_controller.parameters["end_slice"]
            )

        if "remove_small" in self.project_controller.parameters:
            self.extract_histogram_ui.spinBox_3.setValue(
                self.project_controller.parameters["remove_small"]
            )

        if "num_threads" in self.project_controller.parameters:
            self.extract_histogram_ui.spinBox_4.setValue(
                self.project_controller.parameters["num_threads"]
            )

        if "inertia" in self.project_controller.parameters:
            self.extract_histogram_ui.checkBox_34.setChecked(
                self.project_controller.parameters["inertia"]
            )

        if "ferets" in self.project_controller.parameters:
            self.extract_histogram_ui.checkBox_35.setChecked(
                self.project_controller.parameters["ferets"]
            )

        if "entropy" in self.project_controller.parameters:
            self.extract_histogram_ui.checkBox_27.setChecked(
                self.project_controller.parameters["entropy"]
            )

        if "mean_max" in self.project_controller.parameters:
            self.extract_histogram_ui.checkBox_29.setChecked(
                self.project_controller.parameters["mean_max"]
            )

        if "aspect_ratio" in self.project_controller.parameters:
            self.extract_histogram_ui.checkBox_21.setChecked(
                self.project_controller.parameters["aspect_ratio"]
            )

        if "moments" in self.project_controller.parameters:
            self.extract_histogram_ui.checkBox_33.setChecked(
                self.project_controller.parameters["moments"]
            )

        if "euler" in self.project_controller.parameters:
            self.extract_histogram_ui.checkBox_36.setChecked(
                self.project_controller.parameters["euler"]
            )

        if "solidity" in self.project_controller.parameters:
            self.extract_histogram_ui.checkBox_31.setChecked(
                self.project_controller.parameters["solidity"]
            )

        if "histograms" in self.project_controller.parameters:
            self.extract_histogram_ui.checkBox_37.setChecked(
                self.project_controller.parameters["histograms"]
            )

        if "basic_properties" in self.project_controller.parameters:
            self.extract_histogram_ui.checkBox_38.setChecked(
                self.project_controller.parameters["basic_properties"]
            )

        if "select_all" in self.project_controller.parameters:
            self.extract_histogram_ui.checkBox_23.setChecked(
                self.project_controller.parameters["select_all"]
            )

    @property
    def voxel_size(self):
        """Get the voxel size."""
        return self._voxel_size

    @voxel_size.setter
    def voxel_size(self, value):
        """Set the voxel size and update the project parameters."""
        self._voxel_size = value
        self.project_controller.parameters["voxel_size"] = value

    @property
    def background(self):
        """Get the background value."""
        return self._background

    @background.setter
    def background(self, value):
        """Set the background value and update the project parameters."""
        self._background = value
        self.project_controller.parameters["background"] = value

    @property
    def feret_angle(self):
        """Get the feret angle."""
        return self._feret_angle

    @feret_angle.setter
    def feret_angle(self, value):
        """Set the feret angle and update the project parameters."""
        self._feret_angle = value
        self.project_controller.parameters["feret_angle"] = value

    @property
    def mesh_spacing(self):
        """Get the mesh spacing."""
        return self._mesh_spacing

    @mesh_spacing.setter
    def mesh_spacing(self, value):
        """Set the mesh spacing and update the project parameters."""
        self._mesh_spacing = value
        self.project_controller.parameters["mesh_spacing"] = value

    @property
    def start_slice(self):
        """Get the start slice."""
        return self._start_slice

    @start_slice.setter
    def start_slice(self, value):
        """Set the start slice and update the project parameters."""
        self._start_slice = value
        self.project_controller.parameters["start_slice"] = value

    @property
    def end_slice(self):
        """Get the end slice."""
        return self._end_slice

    @end_slice.setter
    def end_slice(self, value):
        """Set the end slice and update the project parameters."""
        self._end_slice = value
        self.project_controller.parameters["end_slice"] = value

    @property
    def remove_small(self):
        """Get the remove small value."""
        return self._remove_small

    @remove_small.setter
    def remove_small(self, value):
        """Set the remove small value and update the project parameters."""
        self._remove_small = value
        self.project_controller.parameters["remove_small"] = value

    @property
    def num_threads(self):
        """Get the number of threads."""
        return self._num_threads

    @num_threads.setter
    def num_threads(self, value):
        """Set the number of threads and update the project parameters."""
        self._num_threads = value
        self.project_controller.parameters["num_threads"] = value

    @property
    def inertia(self):
        """Get the inertia property."""
        return self._inertia

    @inertia.setter
    def inertia(self, value):
        """Set the inertia property and update the project parameters."""
        self._inertia = value
        self.project_controller.parameters["inertia"] = value

    @property
    def ferets(self):
        """Get the ferets property."""
        return self._ferets

    @ferets.setter
    def ferets(self, value):
        """Set the ferets property and update the project parameters."""
        self._ferets = value
        self.project_controller.parameters["ferets"] = value

    @property
    def entropy(self):
        """Get the entropy property."""
        return self._entropy

    @entropy.setter
    def entropy(self, value):
        """Set the entropy property and update the project parameters."""
        self._entropy = value
        self.project_controller.parameters["entropy"] = value

    @property
    def mean_max(self):
        """Get the mean max property."""
        return self._mean_max

    @mean_max.setter
    def mean_max(self, value):
        """Set the mean max property and update the project parameters."""
        self._mean_max = value
        self.project_controller.parameters["mean_max"] = value

    @property
    def aspect_ratio(self):
        """Get the aspect ratio property."""
        return self._aspect_ratio

    @aspect_ratio.setter
    def aspect_ratio(self, value):
        """Set the aspect ratio property and update the project parameters."""
        self._aspect_ratio = value
        self.project_controller.parameters["aspect_ratio"] = value

    @property
    def moments(self):
        """Get the moments property."""
        return self._moments

    @moments.setter
    def moments(self, value):
        """Set the moments property and update the project parameters."""
        self._moments = value
        self.project_controller.parameters["moments"] = value

    @property
    def euler(self):
        """Get the Euler number property."""
        return self._euler

    @euler.setter
    def euler(self, value):
        """Set the Euler number property and update the project parameters."""
        self._euler = value
        self.project_controller.parameters["euler"] = value

    @property
    def solidity(self):
        """Get the solidity property."""
        return self._solidity

    @solidity.setter
    def solidity(self, value):
        """Set the solidity property and update the project parameters."""
        self._solidity = value
        self.project_controller.parameters["solidity"] = value

    @property
    def histograms(self):
        """Get the histograms property."""
        return self._histograms

    @histograms.setter
    def histograms(self, value):
        """Set the histograms property and update the project parameters."""
        self._histograms = value
        self.project_controller.parameters["histograms"] = value

    @property
    def basic_properties(self):
        """Get the basic properties flag."""
        return self._basic_properties

    @basic_properties.setter
    def basic_properties(self, value):
        """Set the basic properties flag and update the project parameters."""
        self._basic_properties = value
        self.project_controller.parameters["basic_properties"] = value

    @property
    def select_all(self):
        """Get the select all flag."""
        return self._select_all

    @select_all.setter
    def select_all(self, value):
        """Set the select all flag and update the project parameters."""
        self._select_all = value
        self.project_controller.parameters["select_all"] = value

    # ## Inputs Changes ###

    def on_voxel_size_change(self, value):
        """Handle changes to the voxel size input."""
        self.voxel_size = value

    def on_background_change(self, value):
        """Handle changes to the background input."""
        self.background = value

    def on_feret_angle_change(self, value):
        """Handle changes to the feret angle input."""
        self.feret_angle = value

    def on_mesh_spacing_change(self, value):
        """Handle changes to the mesh spacing input."""
        self.mesh_spacing = value

    def on_start_slice_change(self, value):
        """Handle changes to the start slice input."""
        self.start_slice = value

    def on_end_slice_change(self, value):
        """Handle changes to the end slice input."""
        self.end_slice = value

    def on_remove_small_change(self, value):
        """Handle changes to the remove small input."""
        self.remove_small = value

    def on_num_threads_change(self, value):
        """Handle changes to the number of threads input."""
        self.num_threads = value

    def _check_input(self):
        if self.project_controller.project_path is None:
            return False

        return True

    # ### Input Button Clicks ####
    def on_extract_properties_click(self):
        """Handle the click event for the extract properties button."""
        if not self._check_input():
            self.logger.log.error("No Project opened! Cannot extract properties!")
            if self.interactive:
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Icon.Critical)
                msg.setText("Please open a project first!")
                msg.setWindowTitle("Error")
                msg.exec()
            return

        self.logger.log.info("Extracting properties...")

        if (
            self._start_slice != -1
            and self._end_slice == -1
            or self._start_slice == -1
            and self._end_slice != -1
        ):
            self.logger.log.error("Both start and end slice must be set")
            if self.interactive:
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Icon.Critical)
                msg.setText("Both start and end slice must be set")
                msg.setWindowTitle("Error")
                msg.exec()
            return

        # build the arguments for the solution
        args_dict = {
            "path": self.project_controller.project_path,
            "Size_threshold": self._remove_small,
            "numberTreads": self._num_threads,
            "Stepsize": self._mesh_spacing,
            # "voxel_spacing": self._mesh_spacing,
            "Angle_spacing": self._feret_angle,
            "Voxel_size": self._voxel_size,
            "Background_mean": self._background,
            "start_slice": self._start_slice,
            "end_slice": self._end_slice,
            "properties_list": "label,area,min_intensity,max_intensity,equivalent_diameter,mean_intensity,bbox,centroid",  # noqa: E501
        }

        argv = [str(os.path.dirname(os.path.realpath(__file__)))]

        # iterate through dict and add key value pairs to argv
        for key, value in args_dict.items():
            argv.append(f"--{key}={value}")

        # build the arguments for the solution
        self.logger.log.debug("Creating solution task...")
        solution = "de.mdc:feature_extraction:0.1.0"
        self.task = RunSolutionTask(self.album_api, solution, argv)

        # Connect the error signal to the handle_error method
        self.task.on_error = self.handle_error

        # connect the finished signal to the handle_segmentation_result method
        self.task.on_finished = self.handle_properties_result

        self.logger.log.debug("Running task...")

        if self.interactive:
            self.extract_histogram_ui.pushButton_2.setEnabled(False)

        QThreadPool().globalInstance().start(self.task)

    def handle_properties_result(self):
        """Handle the result of the properties extraction task."""
        self.logger.log.info("Properties extraction finished")
        self.project_controller.save_parameters()

        if self.interactive:
            self.extract_histogram_ui.pushButton_2.setEnabled(True)

        self.task = None

        # create results file for snakemake
        output_file = self.project_controller.project_files_path.joinpath(
            "histogram_created.txt"
        )
        with open(output_file, "w") as f:
            f.write(
                f"Segmentation extracted histograms at {self.project_controller.analysis_path}\n"
            )
            f.write(f"Active Zarr file: {self.project_controller.active_zarr_file}\n")
            f.write(f"Date: {str(datetime.now().strftime('%Y%m%d_%H%M%S'))}")

    def handle_error(self, e: Exception):
        """Handle errors that occur during the extraction process."""
        self.logger.log.error(
            f"An error occurred. See the log for further information: {str(e)}"
        )
        if self.interactive:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Icon.Critical)
            msg.setText("An error occurred")
            msg.setInformativeText(
                "An error occurred. See the log for further information."
            )
            msg.setWindowTitle("Error")
            msg.exec()

            # enable the button again
            self.extract_histogram_ui.pushButton_2.setEnabled(True)

    def on_cancel_click(self):
        """Handle the click event for the cancel button."""
        self.extract_histogram.close()

    # ## Property Checkboxes ###

    def on_inertia_change(self, value):
        """Handle changes to the inertia checkbox."""
        self.inertia = value

    def on_ferets_change(self, value):
        """Handle changes to the ferets checkbox."""
        self.ferets = value

    def on_entropy_change(self, value):
        """Handle changes to the entropy checkbox."""
        self.entropy = value

    def on_mean_max_change(self, value):
        """Handle changes to the mean max checkbox."""
        self.mean_max = value

    def on_aspect_ratio_change(self, value):
        """Handle changes to the aspect ratio checkbox."""
        self.aspect_ratio = value

    def on_moments_change(self, value):
        """Handle changes to the moments checkbox."""
        self.moments = value

    def on_euler_change(self, value):
        """Handle changes to the Euler number checkbox."""
        self.euler = value

    def on_solidity_change(self, value):
        """Handle changes to the solidity checkbox."""
        self.solidity = value

    # ### Checkbox Clicks ####

    def on_basic_properties_click(self, state):
        """Handle changes to the basic properties checkbox."""
        self.basic_properties = state == QtCore.Qt.CheckState.Checked

    def on_histograms_click(self, state):
        """Handle changes to the histograms checkbox."""
        self.histograms = state == QtCore.Qt.CheckState.Checked

    def on_select_all_click(self, state):
        """Handle changes to the select all checkbox."""
        self.select_all = state
        checkboxes = [
            (self.extract_histogram_ui.checkBox_34, self.on_inertia_change),
            (self.extract_histogram_ui.checkBox_35, self.on_ferets_change),
            (self.extract_histogram_ui.checkBox_27, self.on_entropy_change),
            (self.extract_histogram_ui.checkBox_29, self.on_mean_max_change),
            (self.extract_histogram_ui.checkBox_21, self.on_aspect_ratio_change),
            (self.extract_histogram_ui.checkBox_33, self.on_moments_change),
            (self.extract_histogram_ui.checkBox_36, self.on_euler_change),
            (self.extract_histogram_ui.checkBox_31, self.on_solidity_change),
        ]
        for checkbox, change_function in checkboxes:
            checkbox.setChecked(state)
            change_function(state)

    def connect_buttons(self):
        """Connect the buttons and checkboxes to their respective methods."""
        # ## Name to function mapping
        # input
        # spinBox_5 = voxel_size
        # spinBox_6 = background
        # spinBox_8 = feret_angle
        # spinBox_7 = mesh_spacing
        # spinBox_9 = start_slice
        # spinBox_10 = end_slice
        # spinBox_3 = remove_small
        # spinBox_4 = num_threads

        # checkBox_24 = save_labels
        # checkBox_5 = inclusions

        # output
        # checkBox_34 = inertia
        # checkBox_35 = ferets
        # checkBox_27 = entropy
        # checkBox_29 = mean_max
        # checkBox_21 = aspect_ratio
        # checkBox_33 = moments
        # checkBox_36 = euler
        # checkBox_31 = solidity

        # checkBox_37 = histograms
        # checkBox_38 = basic_properties
        # checkBox_23 = select_all

        # buttons
        # pushButton = subdataset
        # pushButton_2 = all_data
        # pushButton_3 = cancel

        # Input SpinBoxes
        self.extract_histogram_ui.spinBox_5.valueChanged.connect(
            self.on_voxel_size_change
        )
        self.extract_histogram_ui.spinBox_6.valueChanged.connect(
            self.on_background_change
        )
        self.extract_histogram_ui.spinBox_8.valueChanged.connect(
            self.on_feret_angle_change
        )
        self.extract_histogram_ui.spinBox_7.valueChanged.connect(
            self.on_mesh_spacing_change
        )
        self.extract_histogram_ui.spinBox_9.valueChanged.connect(
            self.on_start_slice_change
        )
        self.extract_histogram_ui.spinBox_10.valueChanged.connect(
            self.on_end_slice_change
        )
        self.extract_histogram_ui.spinBox_3.valueChanged.connect(
            self.on_remove_small_change
        )
        self.extract_histogram_ui.spinBox_4.valueChanged.connect(
            self.on_num_threads_change
        )

        # Input Buttons
        # self.extract_histogram_ui.pushButton.clicked.connect(self.on_subdataset_click)
        self.extract_histogram_ui.pushButton_2.clicked.connect(
            self.on_extract_properties_click
        )
        self.extract_histogram_ui.pushButton_3.clicked.connect(self.on_cancel_click)

        # Checkboxes
        self.extract_histogram_ui.checkBox_34.stateChanged.connect(
            self.on_inertia_change
        )
        self.extract_histogram_ui.checkBox_35.stateChanged.connect(
            self.on_ferets_change
        )
        self.extract_histogram_ui.checkBox_27.stateChanged.connect(
            self.on_entropy_change
        )
        self.extract_histogram_ui.checkBox_29.stateChanged.connect(
            self.on_mean_max_change
        )
        self.extract_histogram_ui.checkBox_21.stateChanged.connect(
            self.on_aspect_ratio_change
        )
        self.extract_histogram_ui.checkBox_33.stateChanged.connect(
            self.on_moments_change
        )
        self.extract_histogram_ui.checkBox_36.stateChanged.connect(self.on_euler_change)
        self.extract_histogram_ui.checkBox_31.stateChanged.connect(
            self.on_solidity_change
        )

        # Checkboxes that affect other checkboxes
        self.extract_histogram_ui.checkBox_37.stateChanged.connect(
            self.on_histograms_click
        )
        self.extract_histogram_ui.checkBox_38.stateChanged.connect(
            self.on_basic_properties_click
        )
        self.extract_histogram_ui.checkBox_23.stateChanged.connect(
            self.on_select_all_click
        )



"""Controller for the main window of the ari3d application."""
from pathlib import Path

from PyQt6.QtGui import QPixmap
from PyQt6.QtWidgets import QMainWindow, QMessageBox

from ari3d.gui.ari3d_logging import Ari3dLogger
from ari3d.gui.controller.io.album_c import AlbumController
from ari3d.gui.controller.io.help_c import HelpController
from ari3d.gui.controller.io.project_c import OpenProjectController
from ari3d.gui.controller.io.view_c import ImageViewController
from ari3d.gui.controller.solutions.extract_properties_c import (
    ExtractPropertiesDialogController,
)
from ari3d.gui.controller.solutions.particleSeg_c import (
    UI_ParticleSeg3DAdvancedController,
    UI_ParticleSeg3DPredictController,
)
from ari3d.gui.controller.solutions.streamlit_c import DataViewerDialogController
from ari3d.gui.view.MainWindow_ui import Ui_MainWindow

cur_file_path = Path(__file__).parent


class WindowController(QMainWindow):
    """Controller for the main window of the ari3d application."""

    def __init__(
        self,
    ):
        """Initialize the main window controller with the album API and logger."""
        super().__init__()
        self.album_api = AlbumController()
        self.logger = Ari3dLogger()
        # try to install required solutions
        self.album_api.try_install_required()

        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.setWindowTitle("ARI3D - Analyse Regions in 3D X-Ray CT Images")

        # Set the image of the main window
        self.set_image(
            cur_file_path.parent.parent.joinpath(
                "resources", "images", "logo", "logo_w.png"
            )
        )

        # setup instance Window controller
        self.project_controller = OpenProjectController(self)
        self.image_view_controller = ImageViewController(self, self.project_controller)
        self.particle_seg_dialog_controller = UI_ParticleSeg3DPredictController(
            self, self.album_api, self.project_controller, self.image_view_controller
        )
        self.particle_seg_dialog_advanced_controller = (
            UI_ParticleSeg3DAdvancedController(
                self,
                self.album_api,
                self.project_controller,
                self.image_view_controller,
            )
        )
        self.extract_histogram_controller = ExtractPropertiesDialogController(
            self, self.album_api, self.project_controller
        )
        self.data_viewer_controller = DataViewerDialogController(
            self, self.album_api, self.project_controller
        )
        self.help_controller = HelpController(self)

        self.set_connections()

    def set_image(self, image_path: Path):
        """Set the image for the main window."""
        pixmap = QPixmap(str(image_path))
        self.ui.imageLabel.setPixmap(pixmap)

    def set_connections(self):
        """Set up the connections for the main window actions."""
        # project actions
        self.ui.actionCreate_project.triggered.connect(
            self.project_controller.create_project
        )
        self.ui.actionLoad_project.triggered.connect(
            self.project_controller.open_project
        )
        self.ui.actionQuit.triggered.connect(self.close)

        # 3d viewer actions
        self.ui.actionOpenImage.triggered.connect(self.openImage)
        self.ui.actionClose3DViewer.triggered.connect(self.closeViewer)
        self.ui.actionSegmentation.triggered.connect(self.openParticleSegWindow)

        # process actions
        self.ui.actionOpenDataViewer.triggered.connect(self.openStreamlitWindow)
        self.ui.actionExport_histograms.triggered.connect(self.extractProperties)

        # step actions
        self.ui.actionCheck_steps.triggered.connect(self.check_steps)
        self.ui.actionUpdate_steps.triggered.connect(self.update_steps)
        self.ui.actionReinstall_steps.triggered.connect(self.reinstall_steps)

        # help actions
        self.ui.actionAbout.triggered.connect(self.help_controller.show_about)
        self.ui.actionCode.triggered.connect(self.help_controller.open_code_repository)
        self.ui.actionDocumentation.triggered.connect(
            self.help_controller.open_documentation
        )
        self.ui.actionTutorial.triggered.connect(self.help_controller.open_tutorial)
        self.ui.actionArticle.triggered.connect(self.help_controller.open_article)

        # logging actions
        self.ui.actionLogNone.triggered.connect(self.logger.set_log_level_none)
        self.ui.actionLogDebug.triggered.connect(self.logger.set_log_level_debug)
        self.ui.actionLogInfo.triggered.connect(self.logger.set_log_level_info)
        self.ui.actionLogWarning.triggered.connect(self.logger.set_log_level_warning)

    def check_steps(self):
        """Check for updates of the steps in the album."""
        self.album_api.check_steps()

    def update_steps(self):
        """Update the steps in the album."""
        self.album_api.update_steps()

    def reinstall_steps(self):
        """Reinstall all steps in the album."""
        # ask again for permission
        reply = QMessageBox.question(
            self,
            "Confirm Reinstall",
            "This will re-install all steps. Are you sure?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.No,
        )
        if reply == QMessageBox.StandardButton.Yes:
            self.album_api.reinstall_steps()

    def openImage(self):
        """Open the image viewer window."""
        self.image_view_controller.setup_ui()
        self.image_view_controller.open_file_window.show()

    def closeViewer(self):
        """Close the image viewer window."""
        self.image_view_controller.close_viewer()

    def openParticleSegWindow(self):
        """Open the particle segmentation window."""
        self.particle_seg_dialog_controller.setup_ui()
        # self.particle_seg_dialog_controller.particle_seg_window_ui.pushButton.clicked.connect(self.openParticleSegWindowAdvanced)
        self.particle_seg_dialog_controller.particle_seg_window.show()

    def openParticleSegWindowAdvanced(self):
        """Open the advanced particle segmentation window."""
        self.particle_seg_dialog_controller.particle_seg_window.hide()

        # open the advanced particle segmentation window
        self.particle_seg_dialog_advanced_controller.setup_ui()
        self.particle_seg_dialog_advanced_controller.particle_seg_window.show()

    def openStreamlitWindow(self):
        """Open the streamlit data viewer window."""
        self.data_viewer_controller.setup_ui()
        self.data_viewer_controller.streamlit_window.show()

    def extractProperties(self):
        """Open the extract properties dialog."""
        self.extract_histogram_controller.setup_ui()
        self.extract_histogram_controller.extract_histogram.show()



# importing all the required packages
import altair as alt
import numpy as np
import pandas as pd
import streamlit as st

st.set_page_config(layout="wide", page_title="Step 7 Overview with properties")
tabHistOverview, tabHistogramProperty = st.tabs(
    ["Histogram Overview", "Histograms + Properties"]
)


@st.cache_data(experimental_allow_widgets=True)
def loadHistograms():
    uploadHistoBut = st.sidebar.file_uploader("Upload Histograms", type="csv")
    histogramsRawData = pd.read_csv(uploadHistoBut, encoding="unicode_escape")
    histogramsRawData.index = histogramsRawData.iloc[:, 0]
    histogramsRawData = histogramsRawData.iloc[:, 2:]
    return histogramsRawData


@st.cache_data(experimental_allow_widgets=True)
def loadProperties():
    uploadPropertyBut = st.sidebar.file_uploader("Upload Properties", type="csv")
    propertiesData = pd.read_csv(uploadPropertyBut, encoding="unicode_escape")
    return propertiesData


histogramsData = loadHistograms()
propertiesData = loadProperties()
maxParticleNumber = len(histogramsData)

if "particleLabels" not in st.session_state:
    st.session_state["particleLabels"] = []
if "data_8bit" not in st.session_state:
    st.session_state["data_8bit"] = []
if "plotDataframe" not in st.session_state:
    st.session_state["plotDataframe"] = []
if "Particle_X" not in st.session_state:
    st.session_state["Particle_X"] = 0
if "Particle_A" not in st.session_state:
    st.session_state["Particle_A"] = 0
if "Particle_B" not in st.session_state:
    st.session_state["Particle_B"] = 0
if "Particle_C" not in st.session_state:
    st.session_state["Particle_C"] = 0
if "Particle_D" not in st.session_state:
    st.session_state["Particle_D"] = 0
if "Particle_E" not in st.session_state:
    st.session_state["Particle_E"] = 0
if "Particle_F" not in st.session_state:
    st.session_state["Particle_F"] = 0


@st.cache_data
def histogramsOverview(plotData):
    colorStd = alt.Color(
        "frequency:Q",
        scale=alt.Scale(scheme="viridis", domainMax=0.06),
        legend=alt.Legend(orient="bottom"),
        title="Frequency",
    )
    particleNumber = alt.X("X:N", title="Particle")
    greyBin = alt.Y("Y:O", title="Binned Greyscale").bin(maxbins=52)
    heatMapPartSelect = alt.selection_point(
        encodings=["x"], fields=["X"]
    )  # to select points on a trigger defined in "encodings", e.g. XY position
    opacitySelection = alt.condition(heatMapPartSelect, alt.value(1.0), alt.value(0.2))

    plotAllHistograms = (
        alt.Chart(plotData, width=1600, height=600)
        .mark_area(opacity=0.2)
        .encode(
            x=alt.X("Y", title="Greyscale"),
            y=alt.Y("frequency", title="Frequency"),
            color=(particleNumber),
            tooltip=("X"),
        )
        .transform_filter(heatMapPartSelect)
        .interactive(bind_x=False, bind_y=True)
    )
    heatMapHistograms = (
        alt.Chart(plotData, width=1900, height=400)
        .mark_rect()
        .encode(
            x=particleNumber,
            y=greyBin,
            color=colorStd,
            opacity=opacitySelection,
            tooltip=("X", "Y"),
        )
        .add_params(heatMapPartSelect)
        .interactive()
    )
    plot = heatMapHistograms & plotAllHistograms
    st.altair_chart(plot, use_container_width=True)


binInput = st.sidebar.number_input("bins", value=256, max_value=65535, step=2)


def createSubdata1(n):
    labels_array = np.array(histogramsData.index)
    labels_array = labels_array[labels_array > 0]
    random_labels = np.random.choice(labels_array, n, replace=False)
    random_labels = np.sort(random_labels)
    random_labels = pd.DataFrame(random_labels)
    random_labels = random_labels.rename(
        columns={random_labels.columns[0]: "Label Index"}
    )
    subData1 = random_labels.merge(
        histogramsData, left_on="Label Index", right_on=histogramsData.index, how="left"
    )
    subData1.index = subData1["Label Index"]
    subData1 = subData1.iloc[:, 1:]

    st.session_state["particleLabels"] = random_labels
    return random_labels, subData1


def binningFrom16bit(binInput, bit16):
    bit16toOther = pd.DataFrame(bit16)
    binning = int(65535 / binInput)
    file1 = []
    Index_particle_labels = []
    k = 0
    for index, row in bit16toOther.iterrows():
        if k > -1:
            num = row.to_numpy()
            num = num.ravel()
            rang = int(round(len(num) / binning))
            bins = np.linspace(0, 65535, rang)
            full_range = np.linspace(0, 65534, 65535)
            digitized = np.digitize(full_range, bins)
            bin_sum = [num[digitized == i].sum() for i in range(1, len(bins))]
            file1.append([bin_sum])
            Index_particle_labels.append([index])
        k = k + 1
    x = np.linspace(0, len(bit16toOther.columns) - 1, rang - 1)
    x = x.astype(int)
    file1 = np.array(file1)
    file1 = file1.reshape(file1.shape[0], (file1.shape[1] * file1.shape[2]))
    file1 = pd.DataFrame(file1)
    Index_particle_labels = pd.DataFrame(Index_particle_labels)
    file1.index = Index_particle_labels.iloc[:, 0]
    st.session_state["data_8bit"] = file1
    return file1


def normalizeVolume(unNormalized):
    unNormalized = pd.DataFrame(unNormalized)
    df_new = unNormalized.loc[:, :].div(unNormalized.sum(axis=1), axis=0)
    df_new = df_new.fillna(0)
    return df_new


def transformColumnsXY(subData_ready):
    array = np.array(subData_ready)
    number_of_rows = len(array)
    number_of_columns = len(array[0])
    row = []
    for i in range(number_of_rows):
        for j in range(number_of_columns):
            index_array = (normalizedSubData1_8bit.index)[i]
            element = array[i][j]
            row.append([index_array, j, element])
    DataFrame = pd.DataFrame(row, dtype="float32")
    DataFrame.columns = ["X", "Y", "frequency"]
    pd.set_option("display.float_format", "{:.10f}".format)
    DataFrame["X"] = DataFrame["X"].astype(int)
    DataFrame["Y"] = DataFrame["Y"].astype(int)
    st.session_state["plotDataframe"] = DataFrame


def plotHistogramProperty():
    colorStd2 = alt.Color(
        "X:N",
        scale=alt.Scale(scheme="accent"),
        legend=alt.Legend(title="Particle Number", orient="bottom"),
    )
    colorStd3 = alt.Color(
        propertiesColor,
        scale=alt.Scale(scheme="spectral"),
        legend=alt.Legend(title="Color Property", orient="bottom"),
    )
    colorStd4 = alt.Color("label:N", scale=alt.Scale(scheme="accent"), legend=None)
    sizeStd1 = alt.Size(
        propertiesSize, legend=alt.Legend(title="Size Property", orient="bottom")
    )
    listOfParticles = [
        st.session_state["Particle_X"],
        st.session_state["Particle_A"],
        st.session_state["Particle_B"],
        st.session_state["Particle_C"],
        st.session_state["Particle_D"],
        st.session_state["Particle_E"],
        st.session_state["Particle_F"],
    ]
    plotHist2 = (
        alt.Chart(st.session_state["plotDataframe"], height=1000)
        .mark_line()
        .encode(
            x=alt.X("Y", title="Greyscale"),
            y=alt.Y("frequency", title="Frequency"),
            color=colorStd2,
        )
        .transform_filter(alt.FieldOneOfPredicate(field="X", oneOf=listOfParticles))
        .interactive()
    )
    plotPropSelect = (
        alt.Chart(propertiesData, height=1000)
        .mark_point(filled=True, opacity=1)
        .encode(x=propertiesX, y=propertiesY, size=propertiesSize, color=colorStd4)
        .transform_filter(alt.FieldOneOfPredicate(field="label", oneOf=listOfParticles))
    )
    plotPropAll = (
        alt.Chart(propertiesData)
        .mark_point(shape="triangle", opacity=0.3)
        .encode(x=propertiesX, y=propertiesY, color=colorStd3, size=sizeStd1)
        .interactive()
    )
    plotProp = plotPropAll + plotPropSelect
    with tabHistogramProperty:
        colHist, colProp = st.columns(2)
        with colHist:
            st.altair_chart(plotHist2, use_container_width=True)
        with colProp:
            st.altair_chart(plotProp, use_container_width=True)


numbPartSubData = st.sidebar.number_input(
    "number particles in subset", value=2, max_value=maxParticleNumber, step=1
)
buttNewSubdata = st.sidebar.button("Randomize")
checkboxBin = st.sidebar.checkbox("rebin")
if buttNewSubdata:
    particleLabels, subData1 = createSubdata1(numbPartSubData)
    if checkboxBin:
        subData1_binned = binningFrom16bit(binInput, subData1)
    else:
        subData1_binned = subData1
    normalizedSubData1_8bit = normalizeVolume(subData1_binned)
    transformColumnsXY(normalizedSubData1_8bit)

particleNumberBox = st.sidebar.number_input("Particle Number", step=1)
st.session_state["Particle_X"] = particleNumberBox
dropdown1 = st.sidebar.selectbox(
    label="Particle_A", options=st.session_state["particleLabels"]
)
st.session_state["Particle_A"] = dropdown1
dropdown2 = st.sidebar.selectbox(
    label="Particle_B", options=st.session_state["particleLabels"]
)
st.session_state["Particle_B"] = dropdown2
dropdown3 = st.sidebar.selectbox(
    label="Particle_C", options=st.session_state["particleLabels"]
)
st.session_state["Particle_C"] = dropdown3
dropdown4 = st.sidebar.selectbox(
    label="Particle_D", options=st.session_state["particleLabels"]
)
st.session_state["Particle_D"] = dropdown4
dropdown5 = st.sidebar.selectbox(
    label="Particle_E", options=st.session_state["particleLabels"]
)
st.session_state["Particle_E"] = dropdown5
dropdown6 = st.sidebar.selectbox(
    label="Particle_F", options=st.session_state["particleLabels"]
)
st.session_state["Particle_F"] = dropdown6

with tabHistOverview:
    histogramsOverview(st.session_state["plotDataframe"])
    with st.expander("binned data"):
        column8bit, columnLabels = st.columns(2)
        with columnLabels:
            st.dataframe(st.session_state["particleLabels"], hide_index=False)
        with column8bit:
            st.dataframe(st.session_state["data_8bit"], hide_index=True)

with tabHistogramProperty:
    column1, column2, column3, column4 = st.columns(4)
    with column4:
        propertiesSize = st.selectbox(
            "Size property", propertiesData.columns[:].unique(), index=5
        )
    with column3:
        propertiesColor = st.selectbox(
            "Color property", propertiesData.columns[:].unique(), index=2
        )
    with column2:
        propertiesY = st.selectbox(
            "Y-property", propertiesData.columns[:].unique(), index=4
        )
    with column1:
        propertiesX = st.selectbox(
            "X-property", propertiesData.columns[:].unique(), index=1
        )
    plotHistogramProperty()
    with st.expander("Histograms Subset"):
        column8bit, columnLabels = st.columns(2)
        with columnLabels:
            st.dataframe(st.session_state["particleLabels"], hide_index=False)
        with column8bit:
            st.dataframe(st.session_state["data_8bit"], hide_index=True)



from album.runner.api import get_args, get_package_path, setup

env_file = """name:  data_viewer
channels:
  - conda-forge
dependencies:
  - python=3.9
  - pip
  - pip:
      - pillow
      - numpy
      - pandas
      - altair
      - streamlit
      - anndata
      - tqdm
      - joblib
      - scipy
      - PyYAML
"""


def run():
    args = get_args()

    # importing all the required packages
    import os
    import runpy
    import subprocess
    import sys

    streamlit_script_path = os.path.join(str(get_package_path()), "7_9_quantify6.py")

    # build argv for streamlit
    if args.run_online:
        sys.argv = [
            "streamlit",
            "run",
            streamlit_script_path,
            args.data_path,
            args.report_path,
            args.parameter_yml,
            str(args.run_online),
        ]
        runpy.run_module("streamlit", run_name="__main__")
        # additional code can never be reached!
    else:
        # run streamlit script as standalone script
        subprocess.run(
            [
                "python",
                streamlit_script_path,
                args.data_path,
                args.report_path,
                args.parameter_yml,
                str(args.run_online),
            ],
            check=True,
        )


setup(
    group="de.mdc",
    name="data_viewer",
    version="0.1.0",
    title="Data Viewer for 3D Mineral Quantification",
    description="3D mineral quantification of particulate materials via streamlit",
    solution_creators=["Jose R.A. Godinho", "Jan Philipp Albrecht"],
    cite=[
        {
            "text": "3D mineral quantification of particulate materials with rare earth mineral inclusions: Achieving sub-voxel resolution by considering the partial volume and blurring effect: Shuvam Gupta, Vivian Moutinho, Jose R.A. Godinho, Bradley M. Guy, Jens Gutzmer, Tomography of Materials and Structures, 2025",
            "doi": "10.1016/j.tmater.2025.100050",
        }
    ],
    tags=[
        "interactivity",
        "streamlit",
        "workflow",
        "3D mineral quantification",
        "mspacman",
    ],
    license="MIT",
    documentation=[],
    covers=[],
    album_api_version="0.7.0",
    run=run,
    args=[
        {
            "name": "data_path",
            "type": "string",
            "required": True,
            "description": "Path to the histogram files",
        },
        {
            "name": "report_path",
            "type": "string",
            "required": True,
            "description": "Path used as reporting output",
        },
        {
            "name": "parameter_yml",
            "type": "string",
            "required": True,
            "description": "The path to the parameter json holding default values",
        },
        {
            "name": "run_online",
            "type": "boolean",
            "required": False,
            "default": True,
            "description": "Whether to run online or offline",
        },
    ],
    dependencies={"environment_file": env_file},
)



"""Module for the main window UI of the application."""
# Form implementation generated from reading ui file 'MainWindow.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets


class Ui_MainWindow:
    """UI class for the main window of the application."""

    def setupUi(self, MainWindow):
        """Set up the UI components for the main window."""
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(612, 218)

        # Create a new central widget
        self.centralwidget = QtWidgets.QWidget(parent=MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        MainWindow.setCentralWidget(self.centralwidget)

        # Create a QLabel for displaying the image
        self.imageLabel = QtWidgets.QLabel(parent=self.centralwidget)
        self.imageLabel.setObjectName("imageLabel")

        self.imageLabel.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.imageLabel.setText("Image Placeholder")  # Optional placeholder text

        # Create a new menu bar
        self.menubar = QtWidgets.QMenuBar(parent=MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 612, 22))
        self.menubar.setObjectName("menubar")

        MainWindow.setMenuBar(self.menubar)

        # add tabs for the menu bar
        self._add_project(MainWindow)
        self._add_3d_viewer(MainWindow)
        self._add_data_viewer(MainWindow)
        self._add_segmentation(MainWindow)
        self._add_steps(MainWindow)
        self._add_help(MainWindow)
        self._add_logging(MainWindow)

        self._add_classification(MainWindow)
        self._add_quantification(MainWindow)
        self._add_properties(MainWindow)
        self._add_view(MainWindow)
        self._add_process(MainWindow)
        self._add_extra(MainWindow)

        # add actions to the menubar
        self.menubar.addAction(self.menuProject.menuAction())
        self.menubar.addAction(self.menu3dviewer.menuAction())
        self.menubar.addAction(self.menuProcess.menuAction())
        # self.menubar.addAction(self.menuDataViewer.menuAction())
        # self.menubar.addAction(self.menuSegmentation.menuAction())
        # self.menubar.addAction(self.menuClassification.menuAction())
        # self.menubar.addAction(self.menuQuantification.menuAction())
        # self.menubar.addAction(self.menuProperties.menuAction())
        # self.menubar.addAction(self.menuView.menuAction())
        self.menubar.addAction(self.menuSteps.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        # self.menubar.addAction(self.menuExtra.menuAction())
        self.menubar.addAction(self.menuLog.menuAction())

        # add status bar
        # self._define_statusbar(MainWindow)

        # add toolbar
        # self._define_toolbar(MainWindow)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def _define_statusbar(self, MainWindow):
        # Create a new status bar
        self.statusbar = QtWidgets.QStatusBar(parent=MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

    def _add_data_viewer(self, MainWindow):
        self.menuDataViewer = QtWidgets.QMenu(parent=self.menubar)
        self.menuDataViewer.setObjectName("menuDataViewer")
        self.actionOpenDataViewer = QtGui.QAction(parent=MainWindow)
        self.actionOpenDataViewer.setObjectName("actionOpenDataViewer")
        self.actionCloseDataViewer = QtGui.QAction(parent=MainWindow)
        self.actionCloseDataViewer.setObjectName("actionCloseDataViewer")
        self.menuDataViewer.addAction(self.actionOpenDataViewer)
        self.menuDataViewer.addAction(self.actionCloseDataViewer)

    def _add_3d_viewer(self, MainWindow):
        self.menu3dviewer = QtWidgets.QMenu(parent=self.menubar)
        self.menu3dviewer.setObjectName("menu3dViewer")

        # Create a new QAction for the "Open Image"
        self.actionOpenImage = QtGui.QAction(parent=MainWindow)
        self.actionOpenImage.setObjectName("actionOpenImage")

        # Create a new QAction for the "Close"
        self.actionClose3DViewer = QtGui.QAction(parent=MainWindow)
        self.actionClose3DViewer.setObjectName("action3DViewerClose")

        # Create a new QAction for "segmentation"
        self.actionSegmentation = QtGui.QAction(parent=MainWindow)
        self.actionSegmentation.setObjectName("actionSegmentation")

        self.menu3dviewer.addAction(self.actionOpenImage)
        self.menu3dviewer.addAction(self.actionSegmentation)
        self.menu3dviewer.addAction(self.actionClose3DViewer)

    def _add_extra(self, MainWindow):
        # Create a new menu for "Extra"
        self.menuExtra = QtWidgets.QMenu(parent=self.menubar)
        self.menuExtra.setObjectName("menuExtra")
        # Create a new QAction for the "Histograms16to8bit"
        self.actionHistograms16to8bit = QtGui.QAction(parent=MainWindow)
        self.actionHistograms16to8bit.setObjectName("actionHistograms16to8bit")
        # Create a new QAction for the "NormalizeParticleVolume"
        self.actionNormalizeVolume = QtGui.QAction(parent=MainWindow)
        self.actionNormalizeVolume.setObjectName("actionNormalizeVolume")
        self.menuExtra.addAction(self.actionHistograms16to8bit)
        self.menuExtra.addAction(self.actionNormalizeVolume)

    def _add_view(self, MainWindow):
        # Create a new menu for "View"
        self.menuView = QtWidgets.QMenu(parent=self.menubar)
        self.menuView.setObjectName("menuVextraiew")
        # Create a new QAction for the "Image 2D"
        self.actionImage_2D = QtGui.QAction(parent=MainWindow)
        self.actionImage_2D.setObjectName("actionImage_2D")
        # Create a new QAction for the "Image 3D"
        self.actionImage_3D = QtGui.QAction(parent=MainWindow)
        self.actionImage_3D.setObjectName("actionImage_3D")
        # Create a new QAction for the "Table properties"
        self.actionTable_properties = QtGui.QAction(parent=MainWindow)
        self.actionTable_properties.setObjectName("actionTable_properties")
        # Create a new QAction for the "Table histograms"
        self.actionTable_histograms = QtGui.QAction(parent=MainWindow)
        self.actionTable_histograms.setObjectName("actionTable_histograms")
        # Create a new QAction for the "Console"
        self.actionConsole = QtGui.QAction(parent=MainWindow)
        self.actionConsole.setObjectName("actionConsole")
        # Create a new QAction for the "Plot Properties"
        self.actionPlot_Properties = QtGui.QAction(parent=MainWindow)
        self.actionPlot_Properties.setObjectName("actionPlot_Properties")
        # Create a new QAction for the "Plot Histograms"
        self.actionPlot_Histograms = QtGui.QAction(parent=MainWindow)
        self.actionPlot_Histograms.setObjectName("actionPlot_Histograms")
        self.menuView.addAction(self.actionImage_2D)
        self.menuView.addAction(self.actionImage_3D)
        self.menuView.addAction(self.actionTable_properties)
        self.menuView.addAction(self.actionTable_histograms)
        self.menuView.addAction(self.actionConsole)
        self.menuView.addAction(self.actionPlot_Properties)
        self.menuView.addAction(self.actionPlot_Histograms)

    def _add_properties(self, MainWindow):
        # Create a new menu for "Properties"
        self.menuProperties = QtWidgets.QMenu(parent=self.menubar)
        self.menuProperties.setObjectName("menuProperties")
        # Create a new QAction for the "Mineralogy"
        self.actionMineralogy = QtGui.QAction(parent=MainWindow)
        self.actionMineralogy.setObjectName("actionMineralogy")
        # Create a new QAction for the "Surface"
        self.actionSurface = QtGui.QAction(parent=MainWindow)
        self.actionSurface.setObjectName("actionSurface")
        # Create a new QAction for the "Geometry"
        self.actionGeometry = QtGui.QAction(parent=MainWindow)
        self.actionGeometry.setObjectName("actionGeometry")
        self.menuProperties.addAction(self.actionMineralogy)
        self.menuProperties.addAction(self.actionSurface)
        self.menuProperties.addAction(self.actionGeometry)

    def _add_quantification(self, MainWindow):
        # Create a new menu for "Quantification"
        self.menuQuantification = QtWidgets.QMenu(parent=self.menubar)
        self.menuQuantification.setObjectName("menuQuantification")
        # Create a new QAction for the "Run Quantification"
        self.actionRun_Quantification = QtGui.QAction(parent=MainWindow)
        self.actionRun_Quantification.setObjectName("actionRun_Quantification")
        self.menuQuantification.addAction(self.actionRun_Quantification)

    def _add_help(self, MainWindow):
        # Create a new menu for "Help"
        self.menuHelp = QtWidgets.QMenu(parent=self.menubar)
        self.menuHelp.setObjectName("menuHelp")

        # Create a new QAction for the "About"
        self.actionAbout = QtGui.QAction(parent=MainWindow)
        self.actionAbout.setObjectName("actionAbout")

        # Create a new QAction for the "Code"
        self.actionCode = QtGui.QAction(parent=MainWindow)
        self.actionCode.setObjectName("actionCode")

        # Create a new QAction for the "Documentation"
        self.actionDocumentation = QtGui.QAction(parent=MainWindow)
        self.actionDocumentation.setObjectName("actionDocumentation")

        # Create a new QAction for the "Tutorial"
        self.actionTutorial = QtGui.QAction(parent=MainWindow)
        self.actionTutorial.setObjectName("actionTutorial")

        # Create a new QAction for the "Article"
        self.actionArticle = QtGui.QAction(parent=MainWindow)
        self.actionArticle.setObjectName("actionArticle")

        self.menuHelp.addAction(self.actionAbout)
        self.menuHelp.addAction(self.actionCode)
        self.menuHelp.addAction(self.actionDocumentation)
        self.menuHelp.addAction(self.actionTutorial)
        self.menuHelp.addAction(self.actionArticle)

    def _add_process(self, MainWindow):
        # Create a new menu for "Process"
        self.menuProcess = QtWidgets.QMenu(parent=self.menubar)
        self.menuProcess.setObjectName("menuProcess")
        # Create a new QAction for the "Export histograms"
        self.actionExport_histograms = QtGui.QAction(parent=MainWindow)
        self.actionExport_histograms.setObjectName("actionExport_histograms")

        self.actionOpenDataViewer = QtGui.QAction(parent=MainWindow)
        self.actionOpenDataViewer.setObjectName("actionOpenDataViewer")
        self.actionCloseDataViewer = QtGui.QAction(parent=MainWindow)
        self.actionCloseDataViewer.setObjectName("actionCloseDataViewer")

        self.menuProcess.addAction(self.actionExport_histograms)
        self.menuProcess.addAction(self.actionOpenDataViewer)
        self.menuProcess.addAction(self.actionCloseDataViewer)

    def _add_classification(self, MainWindow):
        # Create a new menu for "Classification"
        self.menuClassification = QtWidgets.QMenu(parent=self.menubar)
        self.menuClassification.setObjectName("menuClassification")
        # Create a new QAction for the "Class List"
        self.actionClass_List = QtGui.QAction(parent=MainWindow)
        self.actionClass_List.setObjectName("actionClass_List")
        # Create a new QAction for the "Class Criteria"
        self.actionClass_Criteria = QtGui.QAction(parent=MainWindow)
        self.actionClass_Criteria.setObjectName("actionClass_Criteria")
        # Create a new QAction for the "Run Classifier"
        self.actionRun_Classifier = QtGui.QAction(parent=MainWindow)
        self.actionRun_Classifier.setObjectName("actionRun_Classifier")
        # Create a new QAction for the "Peak Analysis"
        self.actionPeak_Analysis = QtGui.QAction(parent=MainWindow)
        self.actionPeak_Analysis.setObjectName("actionPeak_Analysis")
        self.menuClassification.addAction(self.actionClass_List)
        self.menuClassification.addAction(self.actionClass_Criteria)
        self.menuClassification.addAction(self.actionRun_Classifier)
        self.menuClassification.addAction(self.actionPeak_Analysis)

    def _add_project(self, MainWindow):
        # Create a new menu for "File"
        self.menuProject = QtWidgets.QMenu(parent=self.menubar)
        self.menuProject.setObjectName("menuFile")

        # Create a new QAction for the "Load project"
        self.actionLoad_project = QtGui.QAction(parent=MainWindow)
        self.actionLoad_project.setObjectName("actionLoad_project")

        # Create a new QAction for the "Create project"
        self.actionCreate_project = QtGui.QAction(parent=MainWindow)
        self.actionCreate_project.setObjectName("actionCreate_project")

        # Create a new QAction for the "Quit"
        self.actionQuit = QtGui.QAction(parent=MainWindow)
        self.actionQuit.setObjectName("actionQuit")

        # add actions to the menu file
        self.menuProject.addAction(self.actionCreate_project)
        self.menuProject.addAction(self.actionLoad_project)
        self.menuProject.addAction(self.actionQuit)

    def _add_segmentation(self, MainWindow):
        # Create a new menu for "Segmentation"
        self.menuSegmentation = QtWidgets.QMenu(parent=self.menubar)
        self.menuSegmentation.setObjectName("menuSegmentation")

        # Create a new QAction for the "Pre-process Grey"
        self.actionPre_process_Grey = QtGui.QAction(parent=MainWindow)
        self.actionPre_process_Grey.setObjectName("actionPre_process_Grey")

        # Create a new QAction for the "ParticleSeg3D"
        self.actionParticleSeg3D = QtGui.QAction(parent=MainWindow)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(
            QtGui.QPixmap(
                "c:\\Users\\daass73\\Desktop\\PlayGUI\\Main GUI\\1109play\\../SegmentIcon.png"
            ),
            QtGui.QIcon.Mode.Normal,
            QtGui.QIcon.State.Off,
        )
        self.actionParticleSeg3D.setIcon(icon2)
        self.actionParticleSeg3D.setObjectName("actionParticleSeg3D")

        # Create a new QAction for the "Threshold"
        self.actionThreshold = QtGui.QAction(parent=MainWindow)
        self.actionThreshold.setObjectName("actionThreshold")

        # Create a new QAction for the "Post-Process Binary"
        self.actionPost_Process_Binary = QtGui.QAction(parent=MainWindow)
        self.actionPost_Process_Binary.setObjectName("actionPost_Process_Binary")

        # Add actions to the menu segmentation
        self.menuSegmentation.addAction(self.actionPre_process_Grey)
        self.menuSegmentation.addAction(self.actionParticleSeg3D)
        self.menuSegmentation.addAction(self.actionThreshold)
        self.menuSegmentation.addAction(self.actionPost_Process_Binary)

    def _add_steps(self, MainWindow):
        # Create a new menu for "Segmentation"
        self.menuSteps = QtWidgets.QMenu(parent=self.menubar)
        self.menuSteps.setObjectName("Steps")

        # Create a new QAction for the "Check new steps"
        self.actionCheck_steps = QtGui.QAction(parent=MainWindow)
        self.actionCheck_steps.setObjectName("actionCheck_steps")

        # Create a new QAction for the "Update new steps"
        self.actionUpdate_steps = QtGui.QAction(parent=MainWindow)
        self.actionUpdate_steps.setObjectName("actionUpdate_steps")

        # Create a new QAction for the "Re-install all steps"
        self.actionReinstall_steps = QtGui.QAction(parent=MainWindow)
        self.actionReinstall_steps.setObjectName("actionReinstall_steps")

        self.menuSteps.addAction(self.actionCheck_steps)
        self.menuSteps.addAction(self.actionUpdate_steps)
        self.menuSteps.addAction(self.actionReinstall_steps)

    def _define_toolbar(self, MainWindow):
        # toolbar
        self.toolBar = QtWidgets.QToolBar(parent=MainWindow)
        self.toolBar.setObjectName("toolBar")
        MainWindow.addToolBar(QtCore.Qt.ToolBarArea.TopToolBarArea, self.toolBar)
        # add actions to the toolbar
        self.toolBar.addAction(self.actionLoad_project)
        self.toolBar.addAction(self.actionParticleSeg3D)

    def _add_logging(self, MainWindow):
        # Create a new menu for "Log"
        self.menuLog = QtWidgets.QMenu(parent=self.menubar)
        self.menuLog.setObjectName("menuLog")
        # Create actions for logging levels
        self.actionLogNone = QtGui.QAction(parent=MainWindow)
        self.actionLogNone.setObjectName("actionLogNone")
        self.actionLogInfo = QtGui.QAction(parent=MainWindow)
        self.actionLogInfo.setObjectName("actionLogInfo")
        self.actionLogDebug = QtGui.QAction(parent=MainWindow)
        self.actionLogDebug.setObjectName("actionLogDebug")
        self.actionLogWarning = QtGui.QAction(parent=MainWindow)
        self.actionLogWarning.setObjectName("actionLogWarning")
        # Add actions to the "Log" menu
        self.menuLog.addAction(self.actionLogNone)
        self.menuLog.addAction(self.actionLogInfo)
        self.menuLog.addAction(self.actionLogDebug)
        self.menuLog.addAction(self.actionLogWarning)

    def retranslateUi(self, MainWindow):
        """Set the text for the UI components."""
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Main_MSPaCMAn"))

        # Set titles for the "Help" menu and its actions
        self.menuHelp.setTitle(_translate("MainWindow", "Help"))
        self.actionAbout.setText(_translate("MainWindow", "About"))
        self.actionCode.setText(_translate("MainWindow", "Code"))
        self.actionDocumentation.setText(_translate("MainWindow", "Documentation"))
        self.actionTutorial.setText(_translate("MainWindow", "Tutorial"))
        self.actionArticle.setText(_translate("MainWindow", "Article"))

        # Set titles for the "Project" menu and its actions
        self.menuProject.setTitle(_translate("MainWindow", "Project"))
        self.actionCreate_project.setText(_translate("MainWindow", "Create project"))
        self.actionLoad_project.setText(_translate("MainWindow", "Load project"))
        self.actionQuit.setText(_translate("MainWindow", "Quit"))

        # Set titles for the "3D Viewer" menu and its actions
        self.menu3dviewer.setTitle(_translate("MainWindow", "3D Viewer"))
        self.actionOpenImage.setText(
            _translate("MainWindow", "Open Image in 3D Viewer")
        )
        self.actionSegmentation.setText(_translate("MainWindow", "Segment Image"))
        self.actionClose3DViewer.setText(_translate("MainWindow", "Close 3D Viewer"))

        # Set titles for the "Data Viewer" menu and its actions
        self.menuDataViewer.setTitle(_translate("MainWindow", "Data Viewer"))
        self.actionOpenDataViewer.setText(_translate("MainWindow", "Open Data Viewer"))
        self.actionCloseDataViewer.setText(
            _translate("MainWindow", "Close Data Viewer")
        )

        # Set titles for the "Classification" menu and its actions
        self.menuClassification.setTitle(_translate("MainWindow", "Classify"))
        self.actionClass_List.setText(_translate("MainWindow", "Class List"))
        self.actionClass_Criteria.setText(_translate("MainWindow", "Class Criteria"))
        self.actionRun_Classifier.setText(_translate("MainWindow", "Run Classifier"))
        self.actionPeak_Analysis.setText(_translate("MainWindow", "Peak Analysis"))

        # Set titles for the "Quantification" menu and its actions
        self.menuQuantification.setTitle(_translate("MainWindow", "Quantify"))
        self.actionRun_Quantification.setText(
            _translate("MainWindow", "Run Quantification")
        )

        # Set titles for the "Properties" menu and its actions
        self.menuProperties.setTitle(_translate("MainWindow", "Properties"))
        self.actionMineralogy.setText(_translate("MainWindow", "Mineralogy"))
        self.actionSurface.setText(_translate("MainWindow", "Surface"))
        self.actionGeometry.setText(_translate("MainWindow", "Geometry"))

        # Set titles for the "View" menu and its actions
        self.menuView.setTitle(_translate("MainWindow", "View"))
        self.actionImage_2D.setText(_translate("MainWindow", "Image 2D"))
        self.actionImage_3D.setText(_translate("MainWindow", "Image 3D"))
        self.actionTable_properties.setText(
            _translate("MainWindow", "Table properties")
        )
        self.actionTable_histograms.setText(
            _translate("MainWindow", "Table histograms")
        )
        self.actionConsole.setText(_translate("MainWindow", "Console"))
        self.actionPlot_Properties.setText(_translate("MainWindow", "Plot Properties"))
        self.actionPlot_Histograms.setText(_translate("MainWindow", "Plot Histograms"))

        # Set titles for the "Process" menu and its actions
        self.menuProcess.setTitle(_translate("MainWindow", "Process"))
        self.actionExport_histograms.setText(
            _translate("MainWindow", "Export histograms")
        )

        # Set titles for the "Extra" menu and its actions
        self.menuExtra.setTitle(_translate("MainWindow", "Extra"))
        self.actionHistograms16to8bit.setText(
            _translate("MainWindow", "Histograms16to8bit")
        )
        self.actionNormalizeVolume.setText(
            _translate("MainWindow", "NormalizeParticleVolume")
        )

        # Set titles for the "Segmentation" menu and its actions
        self.menuSegmentation.setTitle(_translate("MainWindow", "Segment"))
        self.actionPre_process_Grey.setText(
            _translate("MainWindow", "Pre-process Grey")
        )
        self.actionParticleSeg3D.setText(_translate("MainWindow", "ParticleSeg3D"))
        self.actionThreshold.setText(_translate("MainWindow", "Threshold"))
        self.actionPost_Process_Binary.setText(
            _translate("MainWindow", "Post-Process Binary")
        )

        # Set titles for the "Log" menu and its actions
        self.menuLog.setTitle(_translate("MainWindow", "Log"))
        self.actionLogNone.setText(_translate("MainWindow", "None"))
        self.actionLogInfo.setText(_translate("MainWindow", "Info"))
        self.actionLogDebug.setText(_translate("MainWindow", "Debug"))
        self.actionLogWarning.setText(_translate("MainWindow", "Warning"))

        # Set titles for the steps menue and its actions
        self.menuSteps.setTitle(_translate("MainWindow", "Steps"))
        self.actionCheck_steps.setText(_translate("MainWindow", "Check for new Steps"))
        self.actionUpdate_steps.setText(_translate("MainWindow", "Update new Steps"))
        self.actionReinstall_steps.setText(
            _translate("MainWindow", "Reinstall all Steps")
        )

        # toolbar
        # self.toolBar.setWindowTitle(_translate("MainWindow", "toolBar"))



import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from scipy import ndimage
from skimage import morphology
from tqdm import tqdm


def replace_extra_rows_with_zeros(data1, data2):
    # If after erosion the particle disappers a rows of 0s are added

    # Find indices in data1 but not in data2
    missing_indices = data1.index.difference(data2.index)
    # Create a DataFrame with missing indices filled with zeros
    missing_rows = pd.DataFrame(0, index=missing_indices, columns=data2.columns)
    # Append the missing rows to data2
    data2_updated = pd.concat([data2, missing_rows])
    # Sort data2_updated by index to maintain order (optional)
    data2_updated = data2_updated.sort_index()
    return data2_updated


def extract_histograms(labels, gray_scale_thresh, number_treads):
    """Extract histograms for each label."""

    def _get_histogram(label, gst):
        hist, bins = np.histogram(
            gst[labels == label], bins=range(65537)
        )  # full Half-precision range
        return hist, label

    unique_labels = np.unique(labels)
    results = Parallel(n_jobs=number_treads, prefer="threads")(
        delayed(_get_histogram)(label, gray_scale_thresh)
        for label in tqdm(unique_labels)
        if label != 0
    )
    histograms = [result[0] for result in results]
    index = [result[1] for result in results]

    # convert to pandas dataframe
    histograms_df = pd.DataFrame(tqdm(histograms))
    histograms_df.index = index
    return histograms_df


def erosion_based_on_labels(label_mask, surface_properties_mean_intensity):
    """Create an eroded image for each label based on the number of erosions fulfilling the threshold criteria."""
    eroded_images = []

    # calculate how many erosion's have to be maximally performed
    unique_erosions = surface_properties_mean_intensity["no_of_erosions"].unique()

    # loop over the maximal number of erosion's
    for no_of_erosions in unique_erosions:
        # all labels for which the calculated no equals current iteration
        labels_with_erosions = surface_properties_mean_intensity.index[
            surface_properties_mean_intensity["no_of_erosions"] == no_of_erosions
        ].values

        # group mask of label mask of selected indices
        group_mask = np.isin(label_mask, labels_with_erosions)

        # erode selected labels no_of_erosions times
        eroded_mask = ndimage.binary_erosion(group_mask, iterations=no_of_erosions)
        eroded_images.append(
            eroded_mask.astype(np.uint8)
        )  # Convert to uint16 to save memory

    # remove potential overlaps of labels caused by multiple erosions and return
    final_eroded_image = np.sum(eroded_images, axis=0)
    final_eroded_image = (final_eroded_image > 0).astype(np.int8)
    del labels_with_erosions, group_mask, eroded_mask

    return final_eroded_image


def get_unique_labels(labels):
    """Get unique labels not toching the border."""
    # Keeping only the unique labels in labelled image and turing rest to 0
    unique = np.unique(labels)

    # assess labels that touch the border of the volume
    unique1 = np.unique(labels[0])
    unique2 = np.unique(labels[-1])
    unique3 = np.unique(labels[:, 0, :])
    unique4 = np.unique(labels[:, -1, :])
    unique5 = np.unique(labels[:, :, 0])
    unique6 = np.unique(labels[:, :, -1])

    # filter for these
    particles_to_delete = np.concatenate(
        [unique1, unique2, unique3, unique4, unique5, unique6]
    )
    unique_filtered = np.array([i for i in unique if i not in particles_to_delete])

    return np.delete(unique_filtered, np.where(unique_filtered == 0))


def delete_small_particles(labels, binary, gray_scale_volume, size_threshold):
    """Remove the particles smaller than 'Size_threshold' from labels and gray scale volume."""
    binary = np.array(binary, dtype=bool)

    # remove small objects
    binary = morphology.remove_small_objects(binary, size_threshold, connectivity=1)
    binary = binary.astype(int)

    # keep only labels that pass the threshold
    labels = labels * binary
    binary = binary.astype(np.uint8)
    gray_scale_thresholded = binary * gray_scale_volume
    return labels, binary, gray_scale_thresholded


def _process_slice(slice_data, unique_labels):
    slice_data = slice_data.copy()  # Make a copy to ensure it is modifiable
    for label in np.unique(slice_data):
        if label not in unique_labels:
            slice_data[slice_data == label] = 0  # Set label to background
    return slice_data


def filter_mask_image(label_mask, unique_labels, n_jobs=-1):
    """Filter mask image for values not in a unique set of labels."""
    result = Parallel(n_jobs=n_jobs)(
        delayed(_process_slice)(label_mask[:, :, i], unique_labels)
        for i in tqdm(range(label_mask.shape[2]), desc="Processing slices")
    )
    return np.stack(result, axis=2)



import gc
import glob
import os
import time
from tkinter.filedialog import askdirectory

import anndata
import nibabel as nib
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from scipy import ndimage
from skimage import io, measure, morphology
from skimage.measure import marching_cubes, mesh_surface_area
from tqdm import tqdm

startTime = time.time()

##################################################################################
############################### Inputs ###########################################
### link to inputs
# Size threshold in number of voxels, default 1000
Size_threshold = 800
# Defines the number of logical processors used in paralel. -1 is default (uses all)
# Only change if it crashes.
numberTreads = -1
# Stepsize for creating mesh (called mesh size in window - resolution of mesh)
Stepsize = 1
# Voxel spacing for creating mesh (keep constant)
voxel_spacing = (1, 1, 1)
# Enter angle spacing for calculationg Feret dia
Angle_spacing = 10
# voxel size in micron
Voxel_size = 16
Background_mean = 6600
# To load only part of the data
start_slice = None  # If loading all data set to: None
end_slice = None
# prompt to choose the directory. Must have folders with names 'binary','grey','data'
def directory():
    path = askdirectory(title="select folder")
    return path


#################################################################################
############################### Load-Save Paths #################################
### it reads tif, tiff, nii.gz and nii

path = directory()
# Load binary image with particle mask
Binary_image_path = os.path.join(path, r"mask\*")
# Load non binary image (grey-scale 16bit)
Non_binary_image_path = os.path.join(path, r"grey\*")
# save geometrical properties
Path_to_save_geometrical_properties = os.path.join(path, r"data\Properties.csv")
# save inner histograms (inside the particle without the eroded voxels)
Path_to_save_inner_volume_histograms = os.path.join(path, r"data\Inner_histograms.h5ad")
# save outer (surface layers consisting of all voxels eroded) volume histograms
Path_to_save_outer_volume_histograms = os.path.join(path, r"data\Outer_histograms.h5ad")
# save bulk histograms (= Inner + Outer)
Path_to_save_bulk_histogram = os.path.join(path, r"data\Bulk_histograms.h5ad")
# save mesh histograms
Path_to_save_surface_mesh_histograms = os.path.join(
    path, r"data\Surface_histogram.h5ad"
)
# save bulk histograms obtained sfter sobel and smoothening
Path_to_save_bulk_eroded_histogram = os.path.join(path, r"data\Eroded_histograms.h5ad")
# save gradient
Path_to_save_gradient = os.path.join(path, r"data\Gradient.csv")
print("loading from:", path)
###################################################################################
########################### load the images from stacks ###########################

# upload image. Enter the starting slice and end slice. In case of whole sample at once start_slice = 0 and end slice = number of slices -1. If 100 slices then end slice will be 99
# No need to metion tiff or tiff if its 2D stack or 3D image also decides dtype automatically based on number of labels. It can be more than 65535.
def Upload_images(Binary_image_path, start_slice, end_slice):
    # Detect all TIFF files in the provided path
    tiff_files = sorted(
        glob.glob(Binary_image_path + "*.tif") + glob.glob(Binary_image_path + "*.tiff")
    )
    nifti_files = glob.glob(Binary_image_path + "*.nii") + glob.glob(
        Binary_image_path + "*.nii.gz"
    )
    if len(tiff_files) > 1:
        # Multiple TIFF files - treat as 2D stack
        cv_img = []
        for img in tiff_files:
            n = io.imread(img, as_gray=True)
            cv_img.append(n)
        images = np.dstack(cv_img)
        images = np.rollaxis(images, -1)
        del cv_img
        del n
    elif len(tiff_files) == 1:
        # Single TIFF file - treat as 3D image
        images = io.imread(tiff_files[0])
    elif len(nifti_files) == 1:
        # Single NIfTI file - treat as 3D image
        images = nib.load(nifti_files[0]).get_fdata()
    else:
        raise ValueError("No compatible image files found in the specified directory.")
    if start_slice is not None and end_slice is not None:
        if start_slice == end_slice:
            raise ValueError("Start and end slice cannot be the same.")
        if images.ndim == 3:
            if start_slice >= 0 and end_slice < images.shape[0]:
                images = images[start_slice : end_slice + 1, :, :]
            else:
                raise ValueError("Invalid slice range for the 3D image stack.")
        else:
            raise ValueError("Slice range can only be specified for 3D image stacks.")
    # Calculate the maximum grey value
    max_grey_value = np.max(images)
    # Decide which unit type image should be based on the max grey value
    if max_grey_value <= 255:
        # Convert to uint8
        images = images.astype(np.uint8)
    elif max_grey_value <= 65535:
        # Convert to uint16
        images = images.astype(np.uint16)
    else:
        images = images.astype(np.uint32)
    gc.collect()
    return images


# Upload binary image
def Upload_and_process_labelled_image(Binary_image_path):
    labels = Upload_images(Binary_image_path, start_slice, end_slice)
    if (
        len(np.unique(labels)) < 3
    ):  # if mask was exported from Dragonfly the particles have greyvalue 255, from Avizo or ImageJ change to 1.
        labels = measure.label(labels)
    else:
        labels = labels
    binary = np.where(labels > 1, 1, 0)
    binary = binary.astype(np.uint8)
    print("Number of particles before processing", labels.max())
    print("Image size", labels.shape)
    return labels, binary


labels1, binary = Upload_and_process_labelled_image(Binary_image_path)
print("mask loaded")
# Upload non binary image
def Upload_non_binary_images(Non_binary_image_path):
    non_binary = Upload_images(Non_binary_image_path, start_slice, end_slice)
    non_binary = non_binary.astype(np.uint16)
    return non_binary


non_binary = Upload_non_binary_images(Non_binary_image_path)
print("grey loaded")
# Get labels that are not pn the edges of the slice or cut of lables
def get_unique_labels(labels):
    unique = np.unique(labels)
    unique1 = np.unique(labels[0])
    unique2 = np.unique(labels[-1])
    unique3 = np.unique(labels[:, 0, :])
    unique4 = np.unique(labels[:, -1, :])
    unique5 = np.unique(labels[:, :, 0])
    unique6 = np.unique(labels[:, :, -1])
    particles_to_delete = np.concatenate(
        [unique1, unique2, unique3, unique4, unique5, unique6]
    )
    unique_filtered = np.array([i for i in unique if i not in particles_to_delete])
    return np.delete(unique_filtered, np.where(unique_filtered == 0))


unique_labels = get_unique_labels(labels1)
print("image labeled")
###################################################################################
########################### Image processing ######################################
# Remove the particles smaller than 'Size_threshold'
def delete_small_particles(labels, binary, non_binary, Size_threshold):
    binary = np.array(binary, dtype=bool)
    binary = morphology.remove_small_objects(binary, Size_threshold, connectivity=1)
    binary = binary.astype(int)
    labels = labels * binary
    binary = binary.astype(np.uint8)
    li_thresholded = binary * non_binary
    return labels, binary, li_thresholded


labels1, binary, li_thresholded = delete_small_particles(
    labels1, binary, non_binary, Size_threshold
)

# Keeping only the unique labesls in labelled image and turing rest to 0
def process_slice(slice_data, unique_labels):
    slice_data = slice_data.copy()  # Make a copy to ensure it is modifiable
    for label in np.unique(slice_data):
        if label not in unique_labels:
            slice_data[slice_data == label] = 0  # Set label to background
    return slice_data


# parallelization
def process_3d_image(image, unique_labels, n_jobs=-1):
    result = Parallel(n_jobs=n_jobs)(
        delayed(process_slice)(image[:, :, i], unique_labels)
        for i in tqdm(range(image.shape[2]), desc="Processing slices")
    )
    return np.stack(result, axis=2)


# Assuming 'labels' and 'unique_labels' are already defined
labels = process_3d_image(labels1, unique_labels)
del labels1
unique_labels = np.unique(labels)
unique_labels = unique_labels[unique_labels != 0]
print("Number of particles afer processing", len(unique_labels))

# Calculation of surfacearea
def calculate_surface_area(label, labels, spacing=voxel_spacing, step_size=Stepsize):
    non_zero_indices = np.argwhere(labels == label)
    if non_zero_indices.shape[0] < 2:
        return 0
    min_indices = non_zero_indices.min(axis=0) - 2
    max_indices = non_zero_indices.max(axis=0) + 2
    min_indices = np.maximum(min_indices, 0)
    max_indices = np.minimum(max_indices, np.array(labels.shape) - 1)
    isolated_region = labels[
        min_indices[0] : max_indices[0] + 1,
        min_indices[1] : max_indices[1] + 1,
        min_indices[2] : max_indices[2] + 1,
    ]
    label_mask = (isolated_region == label).astype(np.uint8)
    label_mask[label_mask != 1] = 0
    if np.any(np.array(label_mask.shape) < 2) or label_mask.sum() == 0:
        return 0
    verts, faces, _, _ = marching_cubes(
        label_mask, level=0, spacing=spacing, step_size=step_size
    )
    surface_area = mesh_surface_area(verts, faces)
    del isolated_region, verts, faces, non_zero_indices, min_indices, max_indices
    return surface_area


# Parallelizing the process to save time
def calculate_surface_areas(
    labels, unique_labels, spacing=voxel_spacing, step_size=Stepsize
):
    results = Parallel(n_jobs=-1)(
        delayed(calculate_surface_area)(
            label, labels, spacing=spacing, step_size=step_size
        )
        for label in tqdm(
            unique_labels, desc="Processing Labels", total=len(unique_labels)
        )
    )
    df = pd.DataFrame({"label": unique_labels, "Surface Area": results})
    return df


surface_areas_df = calculate_surface_areas(
    labels, unique_labels, spacing=voxel_spacing, step_size=Stepsize
)

# Calculate PVE gradients and eroded_image_1 image after 1 eorion and eroded_image_2 after 2 erosions
def PVE_gradient(binary, li_thresholded, labels, Background_mean):
    surface_properties_list = []
    eroded_image = binary.astype(int)
    eroded_image_1 = None
    eroded_image_2 = None
    for i in range(1, 7):
        eroded_image = morphology.binary_erosion(eroded_image).astype(np.uint8)
        if i == 1:  # Store the image after 1 voxel erosion
            eroded_image_1 = eroded_image.copy()
            eroded_image_1 = eroded_image_1.astype(np.uint8)
        if i == 2:  # Store the image after 2 voxel erosions
            eroded_image_2 = eroded_image.copy()
            eroded_image_2 = eroded_image_2.astype(np.uint8)
        surface_diff = binary - eroded_image
        surface_diff = surface_diff.astype(np.uint8)
        surface_non_binary = li_thresholded * surface_diff
        surface_non_binary = surface_non_binary.astype(np.uint16)
        surface_mesh_properties = pd.DataFrame(
            measure.regionprops_table(
                labels, surface_non_binary, properties=["label", "mean_intensity"]
            )
        ).set_index("label")
        surface_mesh_properties = surface_mesh_properties.rename(
            columns={"mean_intensity": f"mean_intensity{i}"}
        )
        surface_properties_list.append(surface_mesh_properties)
        del surface_diff, surface_non_binary, surface_mesh_properties
        gc.collect()
    surface_properties = pd.concat(surface_properties_list, axis=1)
    cols = [f"mean_intensity{i}" for i in range(1, 7)]
    surface_properties = surface_properties[cols]
    gradient_columns = []
    max_mean_intensity = surface_properties.max(axis=1)
    for i in range(1, 7):
        gradient_col_name = f"Gradient_{i}"
        surface_properties[gradient_col_name] = (
            surface_properties[f"mean_intensity{i}"] - Background_mean
        ) / (max_mean_intensity - Background_mean)
        gradient_columns.append(gradient_col_name)
    return surface_properties, eroded_image_1, eroded_image_2


surface_properties_mean_intensity, eroded_image_1, eroded_image_2 = PVE_gradient(
    binary, li_thresholded, labels, Background_mean
)
surface_properties_mean_intensity["Ratio"] = surface_properties_mean_intensity[
    "Gradient_2"
]
surface_properties_mean_intensity.to_csv(Path_to_save_gradient)

# Getting the number of erosions needed for each particle as a new column. Erossion will be done util gradient is greater than 0.9
def count_erosions(row):
    gradient_cols = row.filter(like="Gradient")
    return sum(1 for value in gradient_cols if value < 0.9 or value > 1)
    del gradient_cols


surface_properties_mean_intensity[
    "no_of_erosions"
] = surface_properties_mean_intensity.apply(count_erosions, axis=1)

# Create image after erosion based on Number od ersions calculated
def erosion_based_on_labels(labels, surface_properties_mean_intensity):
    eroded_images = []
    unique_erosions = surface_properties_mean_intensity["no_of_erosions"].unique()
    for no_of_erosions in unique_erosions:
        labels_with_erosions = surface_properties_mean_intensity.index[
            surface_properties_mean_intensity["no_of_erosions"] == no_of_erosions
        ].values
        labels_with_erosions
        group_mask = np.isin(labels, labels_with_erosions)
        eroded_mask = ndimage.binary_erosion(group_mask, iterations=no_of_erosions)
        eroded_images.append(
            eroded_mask.astype(np.uint8)
        )  # Convert to uint16 to save memory
    final_eroded_image = np.sum(eroded_images, axis=0)
    final_eroded_image = (final_eroded_image > 0).astype(np.int8)
    del labels_with_erosions, group_mask, eroded_mask
    return final_eroded_image


final_eroded_image = erosion_based_on_labels(labels, surface_properties_mean_intensity)

print("processing finished, start exporting histograms")
#################################################################################
############################### HISTOGRAMS ######################################
# BULK HISTOGRAMS
# Get the histogram of each particle # parallelized loop
def extract_histograms(labels, li_thresholded):
    def get_histogram(label, li_thresholded):
        hist, bins = np.histogram(li_thresholded[labels == label], bins=range(65537))
        return hist, label

    unique_labels = np.unique(labels)
    results = Parallel(n_jobs=numberTreads, prefer="threads")(
        delayed(get_histogram)(label, li_thresholded)
        for label in tqdm(unique_labels)
        if label != 0
    )  # tqdm added here
    histograms = [result[0] for result in results]
    index = [result[1] for result in results]
    histograms_df = pd.DataFrame(tqdm(histograms))
    histograms_df.index = index
    return histograms_df


Bulk_histograms = extract_histograms(labels, li_thresholded)

# INNER HISTOGRAM
Inner_volume_labels = final_eroded_image * labels
Inner_volume_histograms = extract_histograms(Inner_volume_labels, li_thresholded)
del Inner_volume_labels, final_eroded_image
# If after erosion the particle disappers a rows of 0s are added
def replace_extra_rows_with_zeros(data1, data2):
    # Find indices in data1 but not in data2
    missing_indices = data1.index.difference(data2.index)
    # Create a DataFrame with missing indices filled with zeros
    missing_rows = pd.DataFrame(0, index=missing_indices, columns=data2.columns)
    # Append the missing rows to data2
    data2_updated = pd.concat([data2, missing_rows])
    # Sort data2_updated by index to maintain order (optional)
    data2_updated = data2_updated.sort_index()
    return data2_updated


Inner_volume_histograms = replace_extra_rows_with_zeros(
    Bulk_histograms, Inner_volume_histograms
)

# OUTTER HISTOGRAMS
Outer_volume_histograms = Bulk_histograms - Inner_volume_histograms
Outer_volume_histograms[Outer_volume_histograms < 0] = 0

# Convert histograms to h5ad and save it
def convert_h5ad(histograms_df, Path_to_save_histograms):
    histograms_df.index = histograms_df.index.astype(str)
    histograms_df.columns = histograms_df.columns.astype(str)
    histograms_df.columns = ["bin_" + str(col) for col in histograms_df.columns]
    obs = pd.DataFrame(index=histograms_df.index)
    var = pd.DataFrame(index=histograms_df.columns)
    adata = anndata.AnnData(X=histograms_df, obs=obs, var=var)
    adata.write(Path_to_save_histograms)
    return histograms_df


Inner_volume_histograms = convert_h5ad(
    Inner_volume_histograms, Path_to_save_inner_volume_histograms
)
Outer_volume_histograms = convert_h5ad(
    Outer_volume_histograms, Path_to_save_outer_volume_histograms
)

# SURFACE MESH HISTOGRAM
# Gets the voxel layer at 1 voxel depth surface mesh and convert/save as h5ad
binary_surface_mesh_eroded2 = eroded_image_1 - eroded_image_2
binary_surface_mesh_eroded2_labels = binary_surface_mesh_eroded2 * labels
del binary_surface_mesh_eroded2
histograms_surface_mesh = extract_histograms(
    binary_surface_mesh_eroded2_labels, li_thresholded
)
histograms_surface_mesh = replace_extra_rows_with_zeros(
    Bulk_histograms, histograms_surface_mesh
)
histograms_surface_mesh = convert_h5ad(
    histograms_surface_mesh, Path_to_save_surface_mesh_histograms
)
Bulk_histograms = convert_h5ad(Bulk_histograms, Path_to_save_bulk_histogram)

##################################################################################################
############################################# PROPERTIES #########################################
# List of properties
# area; #bbox; #bbox_area; #centroid; #convex_image; #coords; #equivalent_diameter; #euler_number; #extent
# feret_diameter_max; #filled_area; #filled_image; #image; #inertia_tensor; #inertia_tensor_eigvals; #intensity_image
# label; #local_centroid; #major_axis_length; #max_intensity; #mean_intensity; #min_intensity; #minor_axis_length
# moments; #moments_central; #moments_normalized; #slice; #solidity; #weighted_centroid; #weighted_local_centroid
# weighted_moments; #weighted_moments_central; #weighted_moments_normalized

# From the list above, add to the list the properties to be calculated. Note: area is actually volume
#### link to properties
Properties = [
    "label",
    "area",
    "min_intensity",
    "max_intensity",
    "equivalent_diameter",
    "mean_intensity",
    "bbox",
    "centroid",
]

# Calculate addl geomtrical properties with ferets
def calculate_properties(
    labels,
    li_thresholded,
    Properties,
    Path_to_save_geometrical_properties,
    unique_labels,
    Angle_spacing,
):
    Bulk_properties = (
        pd.DataFrame(
            measure.regionprops_table(labels, li_thresholded, properties=Properties)
        )
        .set_index("label")
        .rename(columns={"area": "Volume"})
    )
    Properties_Bulk = surface_areas_df.merge(Bulk_properties, how="left", on="label")
    Properties_Bulk["Volume"] = (
        Properties_Bulk["Volume"] * Voxel_size * Voxel_size * Voxel_size
    )
    Properties_Bulk["Surface Area"] = (
        Properties_Bulk["Surface Area"] * Voxel_size * Voxel_size
    )
    Properties_Bulk["equivalent_diameter"] = (
        Properties_Bulk["equivalent_diameter"] * Voxel_size
    )
    Properties_Bulk["Sphericity"] = (
        np.pi ** (1 / 3) * (6 * Properties_Bulk["Volume"]) ** (2 / 3)
    ) / (Properties_Bulk["Surface Area"])

    def process_angles(coords, angle1, angle2, angle3):
        theta1 = np.radians(angle1)
        theta2 = np.radians(angle2)
        theta3 = np.radians(angle3)
        rotation_matrix1 = np.array(
            [
                [np.cos(theta1), -np.sin(theta1), 0],
                [np.sin(theta1), np.cos(theta1), 0],
                [0, 0, 1],
            ]
        )
        rotation_matrix2 = np.array(
            [
                [np.cos(theta2), 0, -np.sin(theta2)],
                [0, 1, 0],
                [np.sin(theta2), 0, np.cos(theta2)],
            ]
        )
        rotation_matrix3 = np.array(
            [
                [1, 0, 0],
                [0, np.cos(theta3), -np.sin(theta3)],
                [0, np.sin(theta3), np.cos(theta3)],
            ]
        )
        rotated_coords = np.dot(coords, rotation_matrix1)
        rotated_coords = np.dot(rotated_coords, rotation_matrix2)
        rotated_coords = np.dot(rotated_coords, rotation_matrix3)
        max_distance = np.max(rotated_coords[:, 0]) - np.min(rotated_coords[:, 0])
        min_distance = np.max(rotated_coords[:, 1]) - np.min(rotated_coords[:, 1])
        depth_distance = np.max(rotated_coords[:, 2]) - np.min(rotated_coords[:, 2])
        return max_distance, min_distance, depth_distance

    def calculate_feret_diameters(label, coords, Angle_spacing):
        max_feret_diameter = 0
        min_feret_diameter = np.inf
        angle_combinations = [
            (angle1, angle2, angle3)
            for angle1 in range(0, 180, Angle_spacing)
            for angle2 in range(0, 180, Angle_spacing)
            for angle3 in range(0, 180, Angle_spacing)
        ]
        results = Parallel(n_jobs=-1)(
            delayed(process_angles)(coords, angle1, angle2, angle3)
            for angle1, angle2, angle3 in tqdm(
                angle_combinations, desc=f"Processing label {label}"
            )
        )
        for max_distance, min_distance, depth_distance in results:
            if max_distance > max_feret_diameter:
                max_feret_diameter = max_distance
            if min_distance < min_feret_diameter:
                min_feret_diameter = min_distance
            if depth_distance < min_feret_diameter:
                min_feret_diameter = depth_distance
        return label, max_feret_diameter, min_feret_diameter

    # Filter region_coords to include only labels in unique_labels
    #### link to ferets
    # region_coords_filtered = [(region.label, region.coords) for region in measure.regionprops(labels, intensity_image=li_thresholded, cache=False, extra_properties=None) if region.label in unique_labels]
    # results = Parallel(n_jobs=-1)(delayed(calculate_feret_diameters)(label, coords,Angle_spacing) for label, coords in tqdm(region_coords_filtered, desc="Calculating Feret diameters"))
    # feret_df = pd.DataFrame(results, columns=['label', 'Max_Feret_Diameter', 'Min_Feret_Diameter']).set_index('label')
    # Properties_Bulk = Properties_Bulk.merge(feret_df, how='left', on='label')
    # Properties_Bulk['Feret_ratio'] = Properties_Bulk['Min_Feret_Diameter'] / Properties_Bulk['Max_Feret_Diameter']
    Properties_Bulk.to_csv(Path_to_save_geometrical_properties)
    return Properties_Bulk


# Call the function with appropriate properties
Properties_Bulk = calculate_properties(
    labels,
    li_thresholded,
    Properties,
    Path_to_save_geometrical_properties,
    unique_labels,
    Angle_spacing,
)

#### link to save labels thick box
# label_image_path = os.path.join(path,'labels\*')
# labels = labels.astype(np.uint16)
# io.imsave(label_image_path\labels.tiff",labels)
finishTime = time.time()
print("Elapsed time:", finishTime - startTime)



"""Module for running solutions in a separate thread using QRunnable."""
from typing import Any, List

from PyQt6.QtCore import QRunnable, pyqtSlot

from ari3d.gui.controller.io.album_c import AlbumController


class RunSolutionTask(QRunnable):
    """QRunnable task for running a solution in a separate thread."""

    def __init__(self, album_api: AlbumController, solution: str, argv: List[Any]):
        """Initialize the task with the album API, solution ID, and arguments."""
        super().__init__()
        self.album_api = album_api
        self.solution = solution
        self.argv = argv
        # self.signals = WorkerSignalsSolution()

    @pyqtSlot()
    def run(self):
        """Run the solution using the album API."""
        try:
            import logging

            from album.core.utils.core_logging import push_active_logger

            push_active_logger(
                logging.getLogger("album")
            )  # we are in a thread. push the logger again

            self.album_api.install_solution(self.solution)
            self.album_api.run(self.solution, argv=self.argv)
            if self.on_finished:
                self.on_finished()
        except Exception as e:
            if self.on_error:
                self.on_error(e)
            else:
                raise e



"""Module for Streamlit dialog UI."""
from PyQt6 import QtCore, QtWidgets


class StreamlitDialog:
    """UI class for the Streamlit dialog."""

    def setupUi(self, Dialog):
        """Set up the UI for the Streamlit dialog."""
        Dialog.setObjectName("Dialog")
        Dialog.resize(600, 200)

        self.gridLayout = QtWidgets.QGridLayout(Dialog)
        self.gridLayout.setObjectName("gridLayout")

        self._add_data_path_section(Dialog)
        self._add_report_path_section(Dialog)
        self._add_parameters_file_section(Dialog)
        self._add_button_box(Dialog)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def _add_data_path_section(self, Dialog):
        label = QtWidgets.QLabel(Dialog)
        label.setText("data_path:")
        self.gridLayout.addWidget(label, 0, 0, 1, 1)

        self.data_pathTextEdit = QtWidgets.QTextEdit(Dialog)
        self.data_pathTextEdit.setReadOnly(True)
        self.gridLayout.addWidget(self.data_pathTextEdit, 0, 1, 1, 1)

        button = QtWidgets.QPushButton(Dialog)
        button.setText("Load Data Path")
        button.setToolTip("Path to the data folder")
        self.gridLayout.addWidget(button, 0, 2, 1, 1)
        self.data_pathButton = button

    def _add_report_path_section(self, Dialog):
        label = QtWidgets.QLabel(Dialog)
        label.setText("report_path:")
        self.gridLayout.addWidget(label, 1, 0, 1, 1)

        self.report_pathTextEdit = QtWidgets.QTextEdit(Dialog)
        self.report_pathTextEdit.setReadOnly(True)
        self.gridLayout.addWidget(self.report_pathTextEdit, 1, 1, 1, 1)

        button = QtWidgets.QPushButton(Dialog)
        button.setText("Load Report Path")
        button.setToolTip(
            "Path to the folder where the quantification report should be written"
        )
        self.gridLayout.addWidget(button, 1, 2, 1, 1)
        self.report_pathButton = button

    def _add_parameters_file_section(self, Dialog):
        label = QtWidgets.QLabel(Dialog)
        label.setText("parameters_yml:")
        self.gridLayout.addWidget(label, 2, 0, 1, 1)

        self.parameters_ymlTextEdit = QtWidgets.QTextEdit(Dialog)
        self.parameters_ymlTextEdit.setReadOnly(True)
        self.gridLayout.addWidget(self.parameters_ymlTextEdit, 2, 1, 1, 1)

        button = QtWidgets.QPushButton(Dialog)
        button.setText("Load Parameters File")
        button.setToolTip("Path to the parameters file")
        self.gridLayout.addWidget(button, 2, 2, 1, 1)
        self.parameters_ymlButton = button

    def _add_button_box(self, Dialog):
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(
            QtWidgets.QDialogButtonBox.StandardButton.Ok
            | QtWidgets.QDialogButtonBox.StandardButton.Cancel
        )
        self.buttonBox.button(QtWidgets.QDialogButtonBox.StandardButton.Ok).setText(
            "Run"
        )
        self.gridLayout.addWidget(self.buttonBox, 3, 0, 1, 3)

    def retranslateUi(self, Dialog):
        """Set the text for the UI elements."""
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Streamlit Parameters"))



"""Controller for managing the particle segmentation solution in the ari3d application."""
import json
import os
from datetime import datetime
from pathlib import Path
from zipfile import ZipFile

import requests
import zarr
from PyQt6 import QtWidgets
from PyQt6.QtCore import Qt, QThreadPool
from PyQt6.QtWidgets import QFileDialog, QMessageBox, QProgressDialog, QWidget

from ari3d.gui.ari3d_logging import Ari3dLogger
from ari3d.gui.controller.io.album_c import AlbumController
from ari3d.gui.controller.io.project_c import OpenProjectController
from ari3d.gui.controller.io.view_c import ImageViewController
from ari3d.gui.model.tasks import RunSolutionTask
from ari3d.gui.view.solutions.particleSegWindow_ui import UI_ParticleSeg3DPredict
from ari3d.gui.view.solutions.particleSegWindow_ui_advanced import (
    UI_ParticleSeg3DAdvanced,
)
from ari3d.utils.conversion import zarr2float
from ari3d.utils.operations.file_operations import copy_folder


def _list_models():
    # create folder if not exists:
    if not os.path.exists(CUR_MODELS_DIR):
        os.makedirs(CUR_MODELS_DIR)

    # get all folders in the models directory
    return [
        f
        for f in os.listdir(CUR_MODELS_DIR)
        if os.path.isdir(os.path.join(CUR_MODELS_DIR, f))
    ]


MAX_INT = 2147483647

CUR_DIR = os.path.dirname(os.path.realpath(__file__))
CUR_MODELS_DIR = Path(CUR_DIR).parent.parent.parent.joinpath("resources", "models")
CUR_MODELS = _list_models()
DEFAULT_MODEL_URL = "https://syncandshare.desy.de/index.php/s/id9D9pkATrFw65s/download/Task310_particle_seg.zip"
SOURCE_SPACING = 0.1
SOURCE_PARTICLE_SIZE = 1.0

SOLUTION_DEFAULTS = {
    "_input": "",  # manually set
    "_output": "",  # commandLinkButton
    "_model": CUR_MODELS_DIR,
    "_name": "",
    "_zscore": "5850.29762143569 7078.294543817302",  # VALUE MODEL TRAINED ON - DO NOT CHANGE - THIS IS NOT A USER INPUT  # noqa
    "_target_particle_size": 60,  # VALUE MODEL TRAINED ON - DO NOT CHANGE - THIS IS NOT A USER INPUT - see source_particle_size for input  # noqa
    "_target_spacing": 0.1,  # VALUE MODEL TRAINED ON - DO NOT CHANGE - THIS IS NOT A USER INPUT - see source_spacing for input  # noqa
    "_fold": "0 1 2 3 4",
    "_batch_size": 6,
    "_processes": 4,
    "_min_rel_particle_size": 0.005,  # VALUE MODEL TRAINED ON - DO NOT CHANGE - THIS IS NOT A USER INPUT
}


class UI_ParticleSeg3DBaseController:
    """Base controller for the ParticleSeg3D solution in the ari3d application."""

    def __init__(
        self,
        parent: QWidget,
        album_api: AlbumController,
        open_project_dialog: OpenProjectController,
        image_view_controller: ImageViewController,
        interactive: bool = True,
    ):
        """Initialize the ParticleSeg3D controller with the necessary parameters."""
        self.parent = parent
        self.logger = Ari3dLogger()
        self.open_project_dialog = open_project_dialog
        self.image_view_controller = image_view_controller
        self.album_api = album_api
        self.interactive = interactive  # whether to show dialogs or not
        self.particle_seg_window = None

        self._particle_seg_input = ""

        # solution arguments
        self._input = SOLUTION_DEFAULTS["_input"]
        self._output = SOLUTION_DEFAULTS["_output"]
        self._model = SOLUTION_DEFAULTS["_model"]
        self._name = SOLUTION_DEFAULTS["_name"]
        self._zscore = SOLUTION_DEFAULTS["_zscore"]
        self._target_particle_size_target_particle_size = SOLUTION_DEFAULTS[
            "_target_particle_size"
        ]
        self._target_spacing = SOLUTION_DEFAULTS["_target_spacing"]
        self._fold = SOLUTION_DEFAULTS["_fold"]
        self._batch_size = SOLUTION_DEFAULTS["_batch_size"]
        self._processes = SOLUTION_DEFAULTS["_processes"]
        self._min_rel_particle_size = SOLUTION_DEFAULTS["_min_rel_particle_size"]
        self._save_tiff = True

        # metadata parameters
        self._source_spacing = None
        self._source_particle_size = None

    def download_default_model(self):
        """Download the default model from the specified URL."""
        response = requests.get(DEFAULT_MODEL_URL, stream=True)
        # to update the download progress bar, we need to know the total
        # size of the file and iterate over the response in chunks
        total_size_in_bytes = int(response.headers.get("content-length", 0))

        # Calculate the scaling factor
        scaling_factor = 1
        if total_size_in_bytes > MAX_INT:
            scaling_factor = total_size_in_bytes // MAX_INT + 1

        # Create a QProgressDialog instance
        if self.interactive:
            progress_dialog = QProgressDialog(
                "Downloading...",
                "Cancel",
                0,
                total_size_in_bytes // scaling_factor,
                self.particle_seg_window,
            )
            progress_dialog.setWindowTitle("Download Progress")
            progress_dialog.setWindowModality(Qt.WindowModality.WindowModal)
            progress_dialog.show()

        bytes_downloaded = 0
        with open(CUR_MODELS_DIR / "Task310_particle_seg.zip", "wb") as f:
            for data in response.iter_content(chunk_size=1024):
                f.write(data)
                bytes_downloaded += len(data)

                # Update the progress dialog
                if self.interactive:
                    progress_dialog.setValue(bytes_downloaded // scaling_factor)

                    # Check if the Cancel button was clicked
                    if progress_dialog.wasCanceled():
                        break
                else:
                    # If not interactive, just log the progress
                    if (
                        bytes_downloaded % (10000 * 1024 * scaling_factor) == 0
                    ):  # log every 100 MB
                        self.logger.log.info(
                            f"Downloaded {bytes_downloaded // scaling_factor} of {total_size_in_bytes // scaling_factor} bytes"  # noqa: E501
                        )

        # Close the progress dialog
        if self.interactive:
            progress_dialog.close()

        # UnZip the model
        with ZipFile(CUR_MODELS_DIR / "Task310_particle_seg.zip", "r") as zip_ref:
            zip_ref.extractall(CUR_MODELS_DIR)
        (CUR_MODELS_DIR / "Task310_particle_seg.zip").unlink()


class UI_ParticleSeg3DPredictController(UI_ParticleSeg3DBaseController):
    """Controller for the ParticleSeg3D prediction solution in the ari3d application."""

    def __init__(
        self,
        parent: QWidget,
        album_api: AlbumController,
        open_project_dialog: OpenProjectController,
        image_view_controller: ImageViewController,
        interactive: bool = True,
    ):
        """Initialize the ParticleSeg3D prediction controller with the necessary parameters."""
        super().__init__(
            parent, album_api, open_project_dialog, image_view_controller, interactive
        )
        self.particle_seg_window_ui = UI_ParticleSeg3DPredict()
        self.task = None

    def setup_ui(self):
        """Set up the UI for the ParticleSeg3D prediction dialog."""
        self.particle_seg_window = QtWidgets.QDialog()
        self.particle_seg_window_ui.setupUi(self.particle_seg_window)
        self.particle_seg_window.setWindowTitle("ParticleSeg3D Parameters")

        # check if model dir empty
        if not CUR_MODELS:
            reply = QMessageBox.question(
                self.particle_seg_window,
                "Model Not Found",
                "No model was found. Do you want to download the default model?"
                " This might take some time but will only be done once?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.Yes,
            )
            if reply == QMessageBox.StandardButton.Yes:
                self.download_default_model()

        self.set_defaults_by_parameters()

        # set UI defaults
        _cur_models_list = CUR_MODELS.copy()
        if self.particle_seg_window_ui.model_name_box.count() > 0:
            _added_items = [
                self.particle_seg_window_ui.model_name_box.itemText(i)
                for i in range(self.particle_seg_window_ui.model_name_box.count())
            ]
            for x in _added_items:
                if x in _cur_models_list:
                    _cur_models_list.remove(x)
        self.particle_seg_window_ui.model_name_box.addItems(
            [os.path.basename(x) for x in _cur_models_list]
        )

        self.link_buttons()

        # non-parameter defaults
        self.set_defaults()

    def set_defaults_by_parameters(self):
        """Set defaults based on the parameters from the open project dialog."""
        if "model_path" in self.open_project_dialog.parameters:
            if self.open_project_dialog.parameters["model_path"] in CUR_MODELS:
                self.particle_seg_window_ui.model_name_box.addItem(
                    self.open_project_dialog.parameters["model_path"]
                )

        if "average_particle_size_mm" in self.open_project_dialog.parameters:
            self.particle_seg_window_ui.particle_size_box.setValue(
                self.open_project_dialog.parameters["average_particle_size_mm"]
            )

        if "voxel_size_mm" in self.open_project_dialog.parameters:
            self.particle_seg_window_ui.voxel_size_box.setValue(
                self.open_project_dialog.parameters["voxel_size_mm"]
            )

        if "save_seg_labels" in self.open_project_dialog.parameters:
            self.particle_seg_window_ui.save_label.setChecked(
                self.open_project_dialog.parameters["save_seg_labels"]
            )

        if "apply_to_slice" in self.open_project_dialog.parameters:
            pass  # todo: implement me

    def set_defaults(self):
        """Set default values for the ParticleSeg3D parameters."""
        if not self.interactive:
            global CUR_MODELS
            # download model
            if len(CUR_MODELS) == 0:
                self.download_default_model()
                CUR_MODELS = _list_models()

        if self.open_project_dialog.active_zarr_file_altered is not None:
            self._particle_seg_input = self.open_project_dialog.active_zarr_file_altered
        elif self.open_project_dialog.active_zarr_file is not None:
            self._particle_seg_input = self.open_project_dialog.active_zarr_file
        else:
            self._particle_seg_input = ""
        self._model = (
            os.path.join(CUR_MODELS_DIR, CUR_MODELS[0]) if len(CUR_MODELS) >= 1 else ""
        )
        self._output = self.open_project_dialog.particle_seg_path

        self._source_spacing = (
            self.open_project_dialog.parameters["voxel_size_mm"]
            if "voxel_size_mm" in self.open_project_dialog.parameters
            else SOURCE_SPACING
        )
        self._source_particle_size = (
            self.open_project_dialog.parameters["average_particle_size_mm"]
            if "average_particle_size_mm" in self.open_project_dialog.parameters
            else SOURCE_PARTICLE_SIZE
        )

        # print debug information
        self.logger.log.debug("Setting defaults for ParticleSeg3D:")
        self.logger.log.debug(f"Input: {self._particle_seg_input}")
        self.logger.log.debug(f"Output: {self._output}")
        self.logger.log.debug(f"Model: {self._model}")
        self.logger.log.debug(f"Source Spacing: {self._source_spacing}")
        self.logger.log.debug(f"Source Particle Size: {self._source_particle_size}")

    def _check_input(self):
        self.logger.log.debug("Checking input...")
        if not self.particle_seg_input:
            if self.interactive:
                QMessageBox.information(
                    self.particle_seg_window,
                    "Information",
                    "Please specify the input image path.",
                )
            self.logger.log.error("No input image path specified!")
            return False

        if self.save_tiff:
            if not self.output:
                if self.interactive:
                    QMessageBox.information(
                        self.particle_seg_window,
                        "Information",
                        "Please specify the output path.",
                    )
                self.logger.log.error("No output path specified!")
                return False

        # check if viewer is open
        if self.interactive:
            if self.image_view_controller.viewer is None:
                QMessageBox.information(
                    self.particle_seg_window,
                    "Information",
                    "Please open an image first.",
                )
                self.logger.log.error("No image viewer open!")
                return False

        if ".particle_seg" in str(self.particle_seg_input):
            if self.interactive:
                QMessageBox.information(
                    self.particle_seg_window,
                    "Information",
                    'Do not specify ".particle_seg" as input folder!',
                )
            self.logger.log.error('Input folder should not contain ".particle_seg"!')
            return False

        return True

    def _prepare_input(self):
        self.logger.log.info("Preparing input...")
        # todo: replace me with link structure in the future. why should we copy stuff?
        # copy input to hidden .particle_seg folder
        if os.path.exists(self.particle_seg_input):
            # ensure float dtype for input
            d = zarr.open(str(self.particle_seg_input), mode="r")
            if d.dtype != float:
                zarr2float(
                    self.particle_seg_input,
                    self.open_project_dialog.particle_seg_image_path,
                )
            else:
                copy_folder(
                    self.particle_seg_input,
                    self.open_project_dialog.particle_seg_image_path,
                )

            # necessary because particle seq expects the input to be in a folder in a certain format
            self._input = self.open_project_dialog.particle_seg_path
        self.logger.log.info("Done preparing input!")

    def _check_output_exists(self):
        # if the output of the segmentation already exists ask whether to re-compute - else load result
        if (
            self.output.joinpath(self.open_project_dialog.project_path.stem)
            .joinpath(self.open_project_dialog.project_path.stem + ".zarr")
            .exists()
        ):
            reply = QMessageBox.question(
                self.particle_seg_window,
                "Output Exists",
                "The output folder already exists. Do you want to use these results?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )
            if reply == QMessageBox.StandardButton.Yes:
                self.handle_particle_seg_result()
                return True
        return False

    def _run_particlseSeg3d_solution(self):
        check = self._check_input()
        if not check:
            return

        if self.interactive:
            if self._check_output_exists():
                self.particle_seg_window.hide()
                return

        self._prepare_input()

        self.logger.log.debug("Building solution arguments...")
        argv = [str(os.path.dirname(os.path.realpath(__file__)))]
        for attr in dir(self):
            if attr.startswith("_") and attr in SOLUTION_DEFAULTS.keys():
                value = getattr(self, attr)
                if value != "":
                    # special case for model:
                    if attr == "_model":
                        value = str(CUR_MODELS_DIR.joinpath(value))
                    argv.append(f"--{attr[1:]}={value}")

        # treat special save_tiff case
        if self.save_tiff:
            argv.append(f"--save_tiff={str(self.open_project_dialog.masks_folder)}")

        self.logger.log.debug("Arguments: %s" % argv)

        metadata_path = os.path.join(Path(self._input), "metadata.json")
        self.logger.log.debug("Metadata path: %s" % metadata_path)

        self.logger.log.info("Creating metadata file...")
        self.create_project_metadata_file(metadata_path)

        self.logger.log.debug("Creating solution task...")
        solution = "de.mdc:particleSeg3D-predict:0.1.0"
        self.task = RunSolutionTask(self.album_api, solution, argv)

        # Connect the error signal to the handle_error method
        self.task.on_error = self.handle_error

        # connect the finished signal to the handle_segmentation_result method
        self.task.on_finished = self.handle_particle_seg_result

        self.logger.log.debug("Running task...")

        if self.interactive:
            self.particle_seg_window_ui.pushButton_2.setEnabled(False)

        QThreadPool().globalInstance().start(self.task)

    def create_project_metadata_file(self, metadata_path: Path):
        """Create a metadata file for the project with the necessary parameters."""
        with open(metadata_path, "w") as json_file:
            json.dump(
                {
                    self._particle_seg_input.stem: {
                        "spacing": self.source_spacing,
                        "particle_size": self.source_particle_size,
                    }
                },
                json_file,
                indent=4,
            )
        self.logger.log.info("Metadata file created.")
        with open(metadata_path) as f:
            self.logger.log.debug("Metadata file content: %s" % json.load(f))

    def handle_error(self, e: Exception):
        """Handle errors that occur during the ParticleSeg3D solution execution."""
        self.logger.log.error(
            f"An error occurred. See the log for further information: {str(e)}"
        )

        if self.interactive:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Icon.Critical)
            msg.setText("An error occurred")
            msg.setInformativeText(
                "An error occurred. See the log for further information."
            )
            msg.setWindowTitle("Error")
            msg.exec()

            # enable the button again
            self.particle_seg_window_ui.pushButton_2.setEnabled(True)

    def handle_particle_seg_result(self):
        """Handle the result of the ParticleSeg3D solution execution."""
        if self.interactive:
            zarr_name = Path(self._particle_seg_input).stem + ".zarr"

            # yes, this is in a subfolder
            out_zarr = Path(self.output).joinpath(
                Path(self._particle_seg_input).stem, zarr_name
            )

            if out_zarr.exists():
                # Load Zarr file in output and treat as mask
                data = zarr.open(str(out_zarr), mode="r")
                self.image_view_controller.overlay_instance_segmentation_result(data)
            else:
                QMessageBox.information(
                    self.particle_seg_window,
                    "Information",
                    "No segmentation result found! Consider running ParticleSeg3D first.",
                )

            # enable the button again
            self.particle_seg_window_ui.pushButton_2.setEnabled(True)

        self.open_project_dialog.save_parameters()

        # create output file for snakemake
        output_file = self.open_project_dialog.project_files_path.joinpath(
            "segmentation_created.txt"
        )
        with open(output_file, "w") as f:
            f.write(
                f"Segmentation created successfully at {self.open_project_dialog.masks_folder}\n"
            )
            f.write(f"Active Zarr file: {self.open_project_dialog.active_zarr_file}\n")
            f.write(f"Date: {str(datetime.now().strftime('%Y%m%d_%H%M%S'))}")

        self.task = None  # reset task

    @property
    def particle_seg_input(self):
        """Get the particle segmentation input path."""
        return self._particle_seg_input

    def _on_input_clicked(self):
        # open input dialog
        directory = QFileDialog.getExistingDirectory(
            self.particle_seg_window, "Select a folder"
        )
        if directory:  # if a directory was selected
            self._particle_seg_input = directory

    # ## output
    @property
    def output(self):
        """Get the output path for the particle segmentation."""
        return self._output

    @output.setter
    def output(self, value):
        """Set the output path for the particle segmentation."""
        self._output = value

    def _on_output_clicked(self):
        # open input dialog
        directory = QFileDialog.getExistingDirectory(
            self.particle_seg_window, "Select a folder"
        )
        if directory:  # if a directory was selected
            self.output = Path(directory)

    # ## model
    @property
    def model(self):
        """Get the model path for the particle segmentation."""
        return self._model

    @model.setter
    def model(self, value):
        """Set the model path for the particle segmentation."""
        self._model = value
        self.open_project_dialog.parameters["model_path"] = value

    def _on_model_changed(self, new_text: str):
        self._model = new_text

    # ## target particle size
    @property
    def target_particle_size(self):
        """Get the target particle size for the particle segmentation."""
        return self._target_particle_size

    @target_particle_size.setter
    def target_particle_size(self, value):
        """Set the target particle size for the particle segmentation."""
        self._target_particle_size = value

    # ## source particle size <- this value is important for prediction
    @property
    def source_particle_size(self):
        """Get the source particle size for the particle segmentation."""
        return self._source_particle_size

    @source_particle_size.setter
    def source_particle_size(self, value):
        """Set the source particle size for the particle segmentation."""
        self._source_particle_size = value
        self.open_project_dialog.parameters["average_particle_size_mm"] = value

    def _on_particle_size_box_changed(self, new_value):
        self.source_particle_size = new_value

    # ## target spacing
    @property
    def target_spacing(self):
        """Get the target spacing for the particle segmentation."""
        return self._target_spacing

    @target_spacing.setter
    def target_spacing(self, value):
        """Set the target spacing for the particle segmentation."""
        self._target_spacing = value

    # ## source spacing <- this value is important for prediction
    @property
    def source_spacing(self):
        """Get the source spacing for the particle segmentation."""
        return self._source_spacing

    @source_spacing.setter
    def source_spacing(self, value):
        """Set the source spacing for the particle segmentation."""
        self._source_spacing = value
        self.open_project_dialog.parameters["voxel_size_mm"] = value

    def _on_source_spacing_changed(self, new_value: str):
        self.source_spacing = float(new_value)

    # ## save tiff
    @property
    def save_tiff(self):
        """Get whether to save the segmentation labels as TIFF."""
        return self._save_tiff

    @save_tiff.setter
    def save_tiff(self, value):
        """Set whether to save the segmentation labels as TIFF."""
        self._save_tiff = value
        self.open_project_dialog.parameters["save_seg_labels"] = value

    def _on_save_label_changed(self, new_value):
        self.save_tiff = new_value

    def link_buttons(self):
        """Link the buttons in the ParticleSeg3D dialog to their respective methods."""
        self.particle_seg_window_ui.save_label_path.clicked.connect(
            self._on_output_clicked
        )
        self.particle_seg_window_ui.input_image_button.clicked.connect(
            self._on_input_clicked
        )
        self.particle_seg_window_ui.save_label.stateChanged.connect(
            self._on_save_label_changed
        )
        self.particle_seg_window_ui.particle_size_box.valueChanged.connect(
            self._on_particle_size_box_changed
        )
        self.particle_seg_window_ui.voxel_size_box.valueChanged.connect(
            self._on_source_spacing_changed
        )
        self.particle_seg_window_ui.model_name_box.currentTextChanged.connect(
            self._on_model_changed
        )

        # OK and Cancel button
        self.particle_seg_window_ui.pushButton_2.clicked.connect(
            self._run_particlseSeg3d_solution
        )
        self.particle_seg_window_ui.pushButton_3.clicked.connect(
            self.particle_seg_window.reject
        )


class UI_ParticleSeg3DAdvancedController(UI_ParticleSeg3DBaseController):
    """Controller for the advanced parameters of the ParticleSeg3D solution in the ari3d application."""

    def __init__(
        self,
        parent: QWidget,
        album_api: AlbumController,
        open_project_dialog: OpenProjectController,
        image_view_controller: ImageViewController,
    ):
        """Initialize the ParticleSeg3D advanced parameters controller with the necessary parameters."""
        super().__init__(parent, album_api, open_project_dialog, image_view_controller)
        self.particle_seg_window_ui = UI_ParticleSeg3DAdvanced()

    def setup_ui(self):
        """Set up the UI for the ParticleSeg3D advanced parameters dialog."""
        self.particle_seg_window_ui.setupUi(self.particle_seg_window)
        self.particle_seg_window.setWindowTitle("ParticleSeg3D Advanced Parameters")

    @property
    def input(self):  # noqa: A003
        """Get the input path for the particle segmentation."""
        # take from open image if it is loaded else from the input field
        if self.open_project_dialog.project_path is not None:
            return self.open_project_dialog.project_path

        return self._input

    @property
    def output(self):
        """Get the output path for the particle segmentation."""
        return self._output

    @property
    def model(self):
        """Get the model path for the particle segmentation."""
        return self._model

    @property
    def name(self):
        """Get the name of the particle segmentation."""
        return self._name

    @property
    def zscore(self):
        """Get the z-score for the particle segmentation."""
        return self._zscore

    @property
    def target_particle_size(self):
        """Get the target particle size for the particle segmentation."""
        return self._target_particle_size

    @property
    def target_spacing(self):
        """Get the target spacing for the particle segmentation."""
        return self._target_spacing

    @property
    def fold(self):
        """Get the fold for the particle segmentation."""
        return self._fold

    @property
    def batch_size(self):
        """Get the batch size for the particle segmentation."""
        return self._batch_size

    @property
    def processes(self):
        """Get the number of processes for the particle segmentation."""
        return self._processes

    @property
    def min_rel_particle_size(self):
        """Get the minimum relative particle size for the particle segmentation."""
        return self._min_rel_particle_size

    @property
    def save_tiff(self):
        """Get whether to save the segmentation labels as TIFF."""
        return self._save_tiff



"""Module for dict operations used in the ARI3D project."""
import json
from typing import List


def get_dict_entries_from_attribute_path(obj, attribute_path: str) -> List:
    """Get all entries from a dict or list by given attribute path."""
    parts = attribute_path.split(".")
    return _get_entries(obj, parts)


def str_to_dict(s):
    """Convert a string representation of a dict to an actual dict."""
    return json.loads(s)


def _get_entries(obj, attribute_paths: List) -> List:
    # case single attribute path
    if len(attribute_paths) == 0:
        if isinstance(obj, list):
            return obj
        else:
            return [obj]

    # extract first attribute
    current_path = attribute_paths[0]

    # if given object is a list itself: recursive call for each list entry
    if isinstance(obj, list):
        res = []
        for el in obj:
            res.extend(_get_entries(el, attribute_paths))
        return res

    # recursive iteration through dict-structure to get attribute
    if isinstance(obj, dict):
        if current_path in obj:
            # remove key
            attribute_paths = attribute_paths.copy()
            # remove current_path from attribute_paths (it will be used as dict-identifier)
            attribute_paths.pop(0)
            return _get_entries(obj[current_path], attribute_paths)
        else:
            return []
